"""
FIXME:
 - lol support late registration and such, this is broken
  - need to cache in the registry somehow, cannot iterate over all ModuleImports each time
   - probably use something like dataclass driver context items or some such

TODO:
 - per-type module imports
"""
import threading
import typing as ta

from .... import reflect as rfl
from ...api.contexts import BaseContext
from ...api.contexts import MarshalFactoryContext
from ...api.contexts import UnmarshalFactoryContext
from ...api.types import Marshaler
from ...api.types import MarshalerFactory
from ...api.types import Unmarshaler
from ...api.types import UnmarshalerFactory
from .api import ModuleImport


FactoryT = ta.TypeVar('FactoryT', bound=MarshalerFactory | UnmarshalerFactory)


##


class _ModuleImportingFactory(ta.Generic[FactoryT]):
    def __init__(
            self,
            fac: FactoryT,
            callback: ta.Callable[[], None] | None = None,
    ) -> None:
        super().__init__()

        self._fac = fac
        self._callback = callback

        self._lock = threading.RLock()
        self._last_mis: ta.Any = None

    def _do_import(self, ctx: BaseContext, mis: ta.Sequence[ModuleImport]) -> None:
        c = 0
        for mi in mis:
            if mi.import_if_necessary():
                c += 1

        if c:
            if self._callback is not None:
                self._callback()

    def _import_if_necessary(self, ctx: BaseContext) -> None:
        if (mis := ctx.configs.get_of(None, ModuleImport)) and mis is not self._last_mis:
            with self._lock:
                if (mis := ctx.configs.get_of(None, ModuleImport)) and mis is not self._last_mis:
                    self._do_import(ctx, mis)
                    self._last_mis = mis


class ModuleImportingMarshalerFactory(_ModuleImportingFactory[MarshalerFactory], MarshalerFactory):
    def make_marshaler(self, ctx: MarshalFactoryContext, rty: rfl.Type) -> ta.Callable[[], Marshaler] | None:
        self._import_if_necessary(ctx)
        return self._fac.make_marshaler(ctx, rty)


class ModuleImportingUnmarshalerFactory(_ModuleImportingFactory[UnmarshalerFactory], UnmarshalerFactory):
    def make_unmarshaler(self, ctx: UnmarshalFactoryContext, rty: rfl.Type) -> ta.Callable[[], Unmarshaler] | None:
        self._import_if_necessary(ctx)
        return self._fac.make_unmarshaler(ctx, rty)



import dataclasses as dc

from ... import lang
from ..impl.api.classes.params import get_class_spec
from ..specs import ClassSpec


##


@dc.dataclass(frozen=True, kw_only=True, eq=False)
class MetaclassSpec(lang.Final):
    confer: frozenset[str] = frozenset()

    final_subclasses: bool = False

    abstract_immediate_subclasses: bool = False


DEFAULT_METACLASS_SPEC = MetaclassSpec()


##


def get_metaclass_spec(obj: type | ClassSpec) -> MetaclassSpec:
    cs: ClassSpec
    if isinstance(obj, type):
        if (cs := get_class_spec(obj)) is None:  # type: ignore[assignment]
            return DEFAULT_METACLASS_SPEC

    elif isinstance(obj, ClassSpec):
        cs = obj

    else:
        raise TypeError(obj)

    return cs.get_last_metadata(MetaclassSpec, DEFAULT_METACLASS_SPEC)



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
import io
import time
import typing as ta

from .direct import DirectByteStreamBufferView
from .errors import NeedMoreDataByteStreamBufferError
from .types import BytesLike
from .types import ByteStreamBuffer
from .types import ByteStreamBufferLike
from .types import ByteStreamBufferView
from .types import MutableByteStreamBuffer


##


class ByteStreamBufferReaderAdapter:
    """
    Adapter: ByteStreamBuffer -> file-like reader methods (`read1`, `read`, `readall`).

    This adapter is policy-driven for how it behaves when insufficient bytes are available. The core buffer is
    intentionally non-blocking; blocking behavior (if desired) must be provided via a `fill` callback that supplies more
    bytes into the buffer.

    `policy`:
      - 'raise': raise NeedMoreDataByteStreamBufferError if fewer than `n` bytes are available (for n >= 0).
      - 'return_partial': return whatever is available (possibly b'') up to `n`.
      - 'block': repeatedly call `fill()` until satisfied or until `fill()` signals EOF.

    `fill`:
      - Callable that writes more bytes into the underlying MutableByteStreamBuffer and returns:
          * True  -> made progress / more data may be available
          * False -> EOF / no more data will arrive
      - Only used when policy == 'block'.

    This adapter exists for interop with legacy code that expects file-like objects, but the core design remains
    independent from `io` and blocking semantics.
    """

    def __init__(
            self,
            buf: ByteStreamBuffer,
            *,
            policy: ta.Literal['raise', 'return_partial', 'block'] = 'raise',
            fill: ta.Optional[ta.Callable[[], bool]] = None,
            block_sleep: ta.Union[ta.Callable[[], None], float, None] = None,
    ) -> None:
        super().__init__()

        self._buf = buf
        self._policy = policy
        self._fill = fill
        self._block_sleep = block_sleep

        if self._policy == 'block' and self._fill is None:
            raise ValueError('policy=block requires fill')

    def _on_block(self) -> None:
        if (bs := self._block_sleep) is None:
            return
        elif callable(bs):
            bs()
        else:
            time.sleep(bs)

    def read1(self, n: int = -1, /) -> bytes:
        return self.read(n)

    def read(self, n: int = -1, /) -> bytes:
        buf = self._buf

        if n is None or n < 0:
            return self.readall()

        if not n:
            return b''

        while True:
            ln = len(buf)
            if ln >= n:
                v = buf.split_to(n)
                return v.tobytes()

            if self._policy == 'return_partial':
                if not ln:
                    return b''
                v = buf.split_to(ln)
                return v.tobytes()

            if self._policy == 'raise':
                raise NeedMoreDataByteStreamBufferError

            # block
            if not ta.cast('ta.Callable[[], bool]', self._fill)():
                # EOF
                if not ln:
                    return b''
                v = buf.split_to(ln)
                return v.tobytes()

            self._on_block()

    def readall(self) -> bytes:
        buf = self._buf
        parts: ta.List[bytes] = []

        while True:
            ln = len(buf)
            if ln:
                v = buf.split_to(ln)
                parts.append(v.tobytes())
                continue

            if self._policy == 'block':
                if not ta.cast('ta.Callable[[], bool]', self._fill)():
                    break
                self._on_block()
                continue

            break

        return b''.join(parts)


class ByteStreamBufferWriterAdapter:
    """
    Adapter: file-like writer sink <- ByteStreamBuffer / bytes-like.

    This is intentionally small and dumb: it exists to bridge into code expecting a `.write(...)` method on an object.

    If given a ByteStreamBufferView-like object, it writes segment-by-segment to avoid materializing copies when the
    sink can accept multiple writes efficiently.
    """

    def __init__(self, f: ta.Any) -> None:
        super().__init__()

        self._f = f

    def write(self, data: ta.Any) -> int:
        f = self._f

        if isinstance(data, (bytes, bytearray, memoryview)):
            b = data.tobytes() if isinstance(data, memoryview) else bytes(data)
            return ta.cast(int, f.write(b))

        if isinstance(data, ByteStreamBufferLike):
            total = 0
            for mv in data.segments():
                total += ta.cast(int, f.write(bytes(mv)))
            return total

        if isinstance(data, ByteStreamBufferView):
            b = data.tobytes()
            return ta.cast(int, f.write(b))

        raise TypeError(data)


##


class BytesIoByteStreamBuffer(MutableByteStreamBuffer):
    """
    ByteStreamBuffer/MutableByteStreamBuffer implementation backed by `io.BytesIO`, using `getbuffer()`.

    This exists primarily for interoperability with code that already produces/consumes `BytesIO`, and to demonstrate
    how `getbuffer()` can expose a non-copying `memoryview` of internal storage.

    Caveat (important):
      - Any exported `memoryview` from `BytesIO.getbuffer()` can pin the BytesIO against resizing.
      - If a caller holds onto a view and we attempt to grow/resize internally, `BytesIO` may raise `BufferError`.

    This backing is therefore best suited for controlled scenarios; it is *not* the recommended default buffer backend
    for applications (segmented/bytearray backends are more predictable).
    """

    def __init__(
            self,
            *,
            compaction_threshold: int = 64 * 1024,
    ) -> None:
        super().__init__()

        self._compaction_threshold = compaction_threshold

        self._bio = io.BytesIO()
        self._rpos = 0

        # reserve/commit state
        self._resv: ta.Optional[bytearray] = None
        self._resv_len = 0

    @property
    def max_size(self) -> ta.Optional[int]:
        return None

    def __len__(self) -> int:
        return self._bio.getbuffer().nbytes - self._rpos

    def peek(self) -> memoryview:
        mv = self._bio.getbuffer()
        return mv[self._rpos:]

    def segments(self) -> ta.Sequence[memoryview]:
        mv = self._bio.getbuffer()
        seg = mv[self._rpos:]
        return (seg,) if len(seg) else ()

    def advance(self, n: int, /) -> None:
        if n < 0 or n > len(self):
            raise ValueError(n)
        self._rpos += n
        # Optional compaction heuristic: if we've consumed a lot, rebuild a smaller BytesIO.
        # This may fail if someone holds a getbuffer() view (BufferError).
        if (
                self._rpos and
                self._rpos >= self._compaction_threshold and
                self._rpos >= (self._bio.getbuffer().nbytes // 2)
        ):
            try:
                remaining = self._bio.getbuffer()[self._rpos:].tobytes()
                self._bio = io.BytesIO(remaining)
                self._rpos = 0
            except BufferError as e:  # noqa
                raise

    def split_to(self, n: int, /) -> ByteStreamBufferView:
        if n < 0 or n > len(self):
            raise ValueError(n)
        mv = self._bio.getbuffer()
        out = mv[self._rpos:self._rpos + n]
        self._rpos += n
        return DirectByteStreamBufferView(out)

    def write(self, data: BytesLike, /) -> None:
        if not data:
            return
        if isinstance(data, memoryview):
            data = data.tobytes()
        try:
            self._bio.seek(0, io.SEEK_END)
            self._bio.write(data)
        except BufferError as e:  # noqa
            raise

    def reserve(self, n: int, /) -> memoryview:
        if n < 0:
            raise ValueError(n)
        if self._resv is not None:
            raise RuntimeError('outstanding reserve')
        b = bytearray(n)
        self._resv = b
        self._resv_len = n
        return memoryview(b)

    def commit(self, n: int, /) -> None:
        if self._resv is None:
            raise RuntimeError('no outstanding reserve')
        if n < 0 or n > self._resv_len:
            raise ValueError(n)
        b = self._resv
        self._resv = None
        self._resv_len = 0
        if n:
            self.write(memoryview(b)[:n])

    def coalesce(self, n: int, /) -> memoryview:
        if n < 0 or n > len(self):
            raise ValueError(n)
        if not n:
            return memoryview(b'')
        # BytesIO is always contiguous, just return the requested slice
        mv = self._bio.getbuffer()
        return mv[self._rpos:self._rpos + n]

    def find(self, sub: bytes, start: int = 0, end: ta.Optional[int] = None) -> int:
        # Normalize start/end to valid range
        s, e, _ = slice(start, end, 1).indices(len(self))
        if e < s:
            e = s

        if not sub:
            return s

        # Convert to bytes for searching (memoryview doesn't have find method)
        b = bytes(self._bio.getbuffer())
        idx = b.find(sub, self._rpos + s, self._rpos + e)
        return (idx - self._rpos) if idx >= 0 else -1

    def rfind(self, sub: bytes, start: int = 0, end: ta.Optional[int] = None) -> int:
        # Normalize start/end to valid range
        s, e, _ = slice(start, end, 1).indices(len(self))
        if e < s:
            e = s

        if not sub:
            return e

        # Convert to bytes for searching (memoryview doesn't have rfind method)
        b = bytes(self._bio.getbuffer())
        idx = b.rfind(sub, self._rpos + s, self._rpos + e)
        return (idx - self._rpos) if idx >= 0 else -1



import typing as ta

from ... import cached
from ... import check
from ... import dataclasses as dc
from ... import lang
from ... import reflect as rfl
from ..binder import bind
from ..elements import Elements
from ..elements import as_elements
from ..inspect import KwargsTarget
from ..keys import Key
from ..keys import as_key
from ..multis import MapBinding
from ..multis import SetBinder
from ..multis import SetBinding
from ..multis import is_map_multi_key
from ..multis import is_set_multi_key
from ..providers import FnProvider
from .id import Id


ItemT = ta.TypeVar('ItemT')


##


def bind_set_entry_const(
        multi_key: ta.Any,
        obj: ta.Any,
        *,
        tag: ta.Any | None = None,
) -> Elements:
    multi_key = as_key(multi_key)
    check.arg(is_set_multi_key(multi_key))

    if tag is None:
        tag = Id(id(obj), tag=multi_key.tag)
    obj_key: Key = Key(type(obj), tag=tag)

    return as_elements(
        bind(obj_key, to_const=obj),
        SetBinding(multi_key, obj_key),
    )


def bind_map_entry_const(
        multi_key: ta.Any,
        map_key: ta.Any,
        obj: ta.Any,
        *,
        tag: ta.Any | None = None,
) -> Elements:
    multi_key = as_key(multi_key)
    check.arg(is_map_multi_key(multi_key))

    if tag is None:
        tag = Id(id(obj), tag=multi_key.tag)
    obj_key: Key = Key(type(obj), tag=tag)

    return as_elements(
        bind(obj_key, to_const=obj),
        MapBinding(multi_key, map_key, obj_key),
    )


##


@ta.final
class ItemsBinderHelper(ta.Generic[ItemT]):
    def __init__(self, items_cls: ta.Any) -> None:
        self._items_cls = items_cls

    @cached.property
    def _item_rty(self) -> rfl.Type:
        rty = check.isinstance(rfl.type_(rfl.get_orig_class(self)), rfl.Generic)
        check.is_(rty.cls, self.__class__)
        return check.single(rty.args)

    @dc.dataclass(frozen=True, eq=False)
    class _ItemsBox:
        vs: ta.Sequence

    @cached.property
    def _items_box(self) -> type[_ItemsBox]:
        if isinstance(item_rty := self._item_rty, type):
            sfx = item_rty.__qualname__
        else:
            sfx = str(item_rty).replace("'", '')

        return lang.new_type(  # noqa
            f'{ItemsBinderHelper._ItemsBox.__qualname__}${sfx}@{id(self):x}',
            (ItemsBinderHelper._ItemsBox,),
            {},
        )

    @cached.property
    def _set_key(self) -> Key:
        return as_key(ta.AbstractSet[self._item_rty])  # type: ignore

    def bind_item_consts(self, *items: ItemT) -> Elements:
        return as_elements(
            bind_set_entry_const(self._set_key, self._items_box(items)),
        )

    @dc.dataclass(frozen=True, eq=False)
    @dc.extra_class_params(repr_id=True)
    class _ItemTag:
        pass

    def bind_item(self, **kwargs: ta.Any) -> Elements:
        tag = ItemsBinderHelper._ItemTag()
        item_key: Key = Key(ta.Any, tag=tag)
        items_box_key: Key = Key(self._items_box, tag=tag)
        return as_elements(
            bind(item_key, **kwargs),
            bind(
                items_box_key,
                to_provider=FnProvider(KwargsTarget.of(
                    lambda v: self._items_box([v]),
                    v=item_key,
                )),
            ),
            SetBinding(self._set_key, items_box_key),
        )

    def bind_items_provider(self, **kwargs: ta.Any) -> Elements:
        return as_elements(
            SetBinder[self._item_rty](),  # type: ignore
            bind(
                self._items_cls,
                to_provider=FnProvider(KwargsTarget.of(
                    lambda s: self._items_cls([v for i in s for v in i.vs]),
                    s=self._set_key,
                )),
                **kwargs,
            ),
        )


items_binder_helper = ItemsBinderHelper



import typing as ta

from ..api.types import MarshalerFactory
from ..api.types import UnmarshalerFactory
from .api import Polymorphism
from .api import TypeTagging
from .api import WrapperTypeTagging
from .marshal import PolymorphismMarshalerFactory
from .unmarshal import PolymorphismUnmarshalerFactory


##


def standard_polymorphism_factories(
        poly: Polymorphism,
        tt: TypeTagging = WrapperTypeTagging(),
) -> ta.Sequence[MarshalerFactory | UnmarshalerFactory]:
    out: list[MarshalerFactory | UnmarshalerFactory] = [
        PolymorphismMarshalerFactory(poly, tt),
        PolymorphismUnmarshalerFactory(poly, tt),
    ]

    return out



"""
TODO:
 - field-configurable coercion
"""
import typing as ta

from ... import dataclasses as dc
from ..api.contexts import MarshalContext
from ..api.contexts import UnmarshalContext
from ..api.types import Marshaler
from ..api.types import Unmarshaler
from ..api.values import Value
from ..factories.typemap import TypeMapMarshalerFactory
from ..factories.typemap import TypeMapUnmarshalerFactory


##


PRIMITIVE_TYPES: tuple[type, ...] = (
    bool,
    int,
    float,
    str,
    bytes,
    type(None),
)


##


@dc.dataclass(frozen=True)
class PrimitiveMarshalerUnmarshaler(Marshaler, Unmarshaler):
    ty: type

    def marshal(self, ctx: MarshalContext, o: ta.Any) -> Value:
        if isinstance(o, self.ty):
            return o  # type: ignore
        if isinstance(o, PRIMITIVE_TYPES):
            return self.ty(o)
        raise TypeError(o)

    def unmarshal(self, ctx: UnmarshalContext, v: Value) -> ta.Any:
        if isinstance(v, self.ty):
            return v
        if isinstance(v, PRIMITIVE_TYPES):
            return self.ty(v)
        raise TypeError(v)


PRIMITIVE_MARSHALER_FACTORY = TypeMapMarshalerFactory({  # noqa
    t: PrimitiveMarshalerUnmarshaler(t) for t in PRIMITIVE_TYPES
})

PRIMITIVE_UNMARSHALER_FACTORY = TypeMapUnmarshalerFactory({  # noqa
    t: PrimitiveMarshalerUnmarshaler(t) for t in PRIMITIVE_TYPES
})



# ruff: noqa: UP007
# @omlish-lite
import abc
import dataclasses as dc
import typing as ta

from ....lite.abstract import Abstract


T = ta.TypeVar('T')

YamlErrorOr = ta.Union['YamlError', T]  # ta.TypeAlias


##


class YamlError(Exception, Abstract):
    @property
    @abc.abstractmethod
    def message(self) -> str:
        raise NotImplementedError


class EofYamlError(YamlError):
    @property
    def message(self) -> str:
        return 'eof'


@dc.dataclass()
class GenericYamlError(YamlError):
    obj: ta.Union[str, Exception]

    @property
    def message(self) -> str:
        if isinstance(self.obj, str):
            return self.obj
        else:
            return str(self.obj)


def yaml_error(obj: ta.Union[YamlError, str, Exception]) -> YamlError:
    if isinstance(obj, YamlError):
        return obj
    elif isinstance(obj, (str, Exception)):
        return GenericYamlError(obj)
    else:
        raise TypeError(obj)



import abc

from ..processing.base import Processor
from ..processing.phases import ProcessorPhase
from ..processing.registry import register_processor_type


##


@register_processor_type(phase=ProcessorPhase.POST_SLOTS)
class UpdateAbstractMethodsProcessor(Processor):
    def process(self, cls: type) -> type:
        abc.update_abstractmethods(cls)  # noqa
        return cls



# ruff: noqa: UP043 UP045
# @omlish-lite
import http.client
import typing as ta
import urllib.error
import urllib.request

from ...io.buffers import ReadableListBuffer
from ...lite.check import check
from ..headers import HttpHeaders
from .base import DEFAULT_ENCODING
from .base import HttpClientContext
from .base import HttpClientError
from .base import HttpRequest
from .sync import HttpClient
from .sync import StreamHttpResponse


##


class UrllibHttpClient(HttpClient):
    def _build_request(self, req: HttpRequest) -> urllib.request.Request:
        d: ta.Any
        if (d := req.data) is not None:
            if isinstance(d, str):
                d = d.encode(DEFAULT_ENCODING)

        # urllib headers are dumb dicts [1], and keys *must* be strings or it will automatically add problematic default
        # headers because it doesn't see string keys in its header dict [2]. frustratingly it has no problem accepting
        # bytes values though [3].
        # [1]: https://github.com/python/cpython/blob/232b303e4ca47892f544294bf42e31dc34f0ec72/Lib/urllib/request.py#L319-L325  # noqa
        # [2]: https://github.com/python/cpython/blob/232b303e4ca47892f544294bf42e31dc34f0ec72/Lib/urllib/request.py#L1276-L1279  # noqa
        # [3]: https://github.com/python/cpython/blob/232b303e4ca47892f544294bf42e31dc34f0ec72/Lib/http/client.py#L1300-L1301  # noqa
        h: dict[str, str] = {}
        if hs := req.headers_:
            for k, vs in hs.items():
                h[k] = check.single(vs)

        return urllib.request.Request(  # noqa
            req.url,
            method=req.method_or_default,
            headers=h,
            data=d,
        )

    def _stream_request(self, ctx: HttpClientContext, req: HttpRequest) -> StreamHttpResponse:
        try:
            resp = urllib.request.urlopen(  # noqa
                self._build_request(req),
                timeout=req.timeout_s,
            )

        except urllib.error.HTTPError as e:
            try:
                return StreamHttpResponse(
                    status=e.code,
                    headers=HttpHeaders(e.headers.items()),
                    request=req,
                    underlying=e,
                    _closer=e.close,
                )

            except Exception:
                e.close()
                raise

        except (urllib.error.URLError, http.client.HTTPException) as e:
            raise HttpClientError from e

        try:
            return StreamHttpResponse(
                status=resp.status,
                headers=HttpHeaders(resp.headers.items()),
                request=req,
                underlying=resp,
                _stream=ReadableListBuffer().new_buffered_reader(resp),
                _closer=resp.close,
            )

        except Exception:  # noqa
            resp.close()
            raise



import typing as ta

import anyio

from ..sleeps import AsyncliteSleeps


##


class AnyioAsyncliteSleeps(AsyncliteSleeps):
    def sleep(self, delay: float) -> ta.Awaitable[None]:
        return anyio.sleep(delay)



import dataclasses as dc


##


@dc.dataclass(frozen=True)
class TextwrapOpts:
    # The maximum width of wrapped lines (unless break_long_words is false).
    width: int = 70  # noqa

    _: dc.KW_ONLY

    # String that will be prepended to the first line of wrapped output. Counts towards the line's width.
    initial_indent: str = ''

    # String that will be prepended to all lines save the first of wrapped output; also counts towards each line's
    # width.
    subsequent_indent: str = ''

    # Expand tabs in input text to spaces before further processing. Each tab will become 0 .. 'tabsize' spaces,
    # depending on its position in its line. If false, each tab is treated as a single character.
    expand_tabs: bool = True

    # Expand tabs in input text to 0 .. 'tabsize' spaces, unless 'expand_tabs' is false.
    tabsize: int = 8

    # Replace all whitespace characters in the input text by spaces after tab expansion. Note that if expand_tabs is
    # false and replace_whitespace is true, every tab will be converted to a single space!
    replace_whitespace: bool = True

    # Ensure that sentence-ending punctuation is always followed by two spaces. Off by default because the algorithm is
    # (unavoidably) imperfect.
    fix_sentence_endings: bool = False

    # Break words longer than 'width'.  If false, those words will not be broken, and some lines might be longer than
    # 'width'.
    break_long_words: bool = True

    # Allow breaking hyphenated words. If true, wrapping will occur preferably on whitespaces and right after hyphens
    # part of compound words.
    break_on_hyphens: bool = True

    # Drop leading and trailing whitespace from lines.
    drop_whitespace: bool = True

    # Truncate wrapped lines.
    max_lines: int | None = None

    # Append to the last line of truncated text.
    placeholder: str = ' [...]'



import abc
import typing as ta

from ... import check
from ... import lang


OpTuple: ta.TypeAlias = tuple['Op', ...]


##


class Op(lang.Abstract, lang.PackageSealed):
    def _match_repr(self) -> str:
        return f'{self.__class__.__name__}@{id(self):x}'


class CompositeOp(Op, lang.Abstract):
    def __init_subclass__(cls, **kwargs: ta.Any) -> None:
        super().__init_subclass__(**kwargs)

        try:
            leaf_op_cls = LeafOp
        except NameError:
            pass
        else:
            check.not_issubclass(cls, leaf_op_cls)

    @property
    @abc.abstractmethod
    def children(self) -> OpTuple:
        raise NotImplementedError

    @abc.abstractmethod
    def replace_children(self, *children: Op) -> Op:
        raise NotImplementedError


class LeafOp(Op, lang.Abstract):
    def __init_subclass__(cls, **kwargs: ta.Any) -> None:
        super().__init_subclass__(**kwargs)

        check.not_issubclass(cls, CompositeOp)



# ruff: noqa: UP045
# @omlish-lite
import asyncio
import typing as ta

from ..events import AsyncliteEvent
from ..events import AsyncliteEvents
from .base import AsyncioAsyncliteApi
from .base import AsyncioAsyncliteObject


##


class AsyncioAsyncliteEvent(AsyncliteEvent, AsyncioAsyncliteObject):
    def __init__(self, u: asyncio.Event) -> None:
        super().__init__()

        self._u = u

    def set(self) -> None:
        return self._u.set()

    def is_set(self) -> bool:
        return self._u.is_set()

    async def wait(self, *, timeout: ta.Optional[float] = None) -> None:
        await self._wait_for(self._u.wait(), timeout=timeout)


class AsyncioAsyncliteEvents(AsyncliteEvents, AsyncioAsyncliteApi):
    def make_event(self) -> AsyncliteEvent:
        return AsyncioAsyncliteEvent(asyncio.Event())



from ... import check
from ... import dataclasses as dc
from ... import lang
from .exprs import CanExpr
from .exprs import Expr
from .exprs import ExprBuilder
from .ops import OpKind


##


class BinaryOp(dc.Frozen, lang.Final, eq=False):
    name: str
    kind: OpKind


class BinaryOps(lang.Namespace):
    EQ = BinaryOp('eq', OpKind.CMP)
    NE = BinaryOp('ne', OpKind.CMP)
    LT = BinaryOp('lt', OpKind.CMP)
    LE = BinaryOp('le', OpKind.CMP)
    GT = BinaryOp('gt', OpKind.CMP)
    GE = BinaryOp('ge', OpKind.CMP)

    ADD = BinaryOp('add', OpKind.ARITH)
    SUB = BinaryOp('sub', OpKind.ARITH)
    MUL = BinaryOp('mul', OpKind.ARITH)
    DIV = BinaryOp('div', OpKind.ARITH)
    MOD = BinaryOp('mod', OpKind.ARITH)

    CONCAT = BinaryOp('concat', OpKind.STR)


class Binary(Expr, lang.Final):
    op: BinaryOp
    l: Expr
    r: Expr


class BinaryBuilder(ExprBuilder):
    def binary(self, op: BinaryOp, *es: CanExpr) -> Expr:
        check.not_empty(es)
        l = self.expr(es[0])
        for r in es[1:]:
            l = Binary(op, l, self.expr(r))
        return l

    #

    def eq(self, *es: CanExpr) -> Expr:
        return self.binary(BinaryOps.EQ, *es)

    def ne(self, *es: CanExpr) -> Expr:
        return self.binary(BinaryOps.NE, *es)

    def lt(self, *es: CanExpr) -> Expr:
        return self.binary(BinaryOps.LT, *es)

    def le(self, *es: CanExpr) -> Expr:
        return self.binary(BinaryOps.LE, *es)

    def gt(self, *es: CanExpr) -> Expr:
        return self.binary(BinaryOps.GT, *es)

    def ge(self, *es: CanExpr) -> Expr:
        return self.binary(BinaryOps.GE, *es)

    #

    def add(self, *es: CanExpr) -> Expr:
        return self.binary(BinaryOps.ADD, *es)

    def sub(self, *es: CanExpr) -> Expr:
        return self.binary(BinaryOps.SUB, *es)

    def mul(self, *es: CanExpr) -> Expr:
        return self.binary(BinaryOps.MUL, *es)

    def div(self, *es: CanExpr) -> Expr:
        return self.binary(BinaryOps.DIV, *es)

    def mod(self, *es: CanExpr) -> Expr:
        return self.binary(BinaryOps.MOD, *es)

    #

    def concat(self, *es: CanExpr) -> Expr:
        return self.binary(BinaryOps.CONCAT, *es)



# ruff: noqa: UP006 UP043 UP045
# @omlish-lite
import threading
import typing as ta

from ..lite.maybes import Maybe


T = ta.TypeVar('T')


##


class SyncOnce:
    def __init__(self) -> None:
        super().__init__()

        self._done = False
        self._lock = threading.Lock()

    def do(self, fn: ta.Callable[[], None]) -> bool:
        if self._done:
            return False
        with self._lock:
            if self._done:
                return False  # type: ignore
            try:
                fn()
            finally:
                self._done = True
            return True


class SyncLazy(ta.Generic[T]):
    def __init__(self) -> None:
        super().__init__()

        self._once = SyncOnce()
        self._v: Maybe[T] = Maybe.empty()

    def peek(self) -> Maybe[T]:
        return self._v

    def set(self, v: T) -> None:
        self._v = Maybe.just(v)

    def get(self, fn: ta.Callable[[], T]) -> T:
        def do():
            self._v = Maybe.just(fn())
        self._once.do(do)
        return self._v.must()


class SyncLazyFn(ta.Generic[T]):
    def __init__(self, fn: ta.Callable[[], T]) -> None:
        super().__init__()

        self._fn = fn
        self._once = SyncOnce()
        self._v: Maybe[T] = Maybe.empty()

    def peek(self) -> Maybe[T]:
        return self._v

    def set(self, v: T) -> None:
        self._v = Maybe.just(v)

    def get(self) -> T:
        def do():
            self._v = Maybe.just(self._fn())
        self._once.do(do)
        return self._v.must()



# ruff: noqa: UP045
# @omlish-lite
import abc
import typing as ta

from ...lite.abstract import Abstract
from .base import AsyncliteApi
from .base import AsyncliteCloseable
from .base import AsyncliteObject


T = ta.TypeVar('T')


##


class AsyncliteQueue(AsyncliteCloseable, AsyncliteObject, Abstract, ta.Generic[T]):
    @abc.abstractmethod
    def qsize(self) -> int:
        raise NotImplementedError

    @abc.abstractmethod
    def empty(self) -> bool:
        raise NotImplementedError

    @abc.abstractmethod
    def full(self) -> bool:
        raise NotImplementedError

    @abc.abstractmethod
    def put(self, item: T, *, timeout: ta.Optional[float] = None) -> ta.Awaitable[None]:
        raise NotImplementedError

    @abc.abstractmethod
    def put_nowait(self, item: T) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    def get(self, *, timeout: ta.Optional[float] = None) -> ta.Awaitable[T]:
        raise NotImplementedError

    @abc.abstractmethod
    def get_nowait(self) -> T:
        raise NotImplementedError


class AsyncliteQueues(AsyncliteApi, Abstract):
    @abc.abstractmethod
    def make_queue(self, maxsize: int = 0) -> AsyncliteQueue:
        raise NotImplementedError



# ruff: noqa: UP006 UP007 UP043 UP045
# @omlish-lite
"""
TODO:
 - overhaul and just coro-ify pyio?
"""
import io
import typing as ta

from ..lite.check import check
from .readers import AsyncBufferedBytesReader
from .readers import AsyncRawBytesReader
from .readers import BufferedBytesReader
from .readers import RawBytesReader


##


class ReadableListBuffer:
    # FIXME: merge with PrependableGeneratorReader
    # FIXME: AND PUSHBACKREADER
    # FIXME: replace this whole thing with ByteStreamBuffers

    def __init__(self) -> None:
        super().__init__()

        self._lst: list[bytes] = []
        self._len = 0

    def __bool__(self) -> ta.NoReturn:
        raise TypeError("Use 'buf is not None' or 'len(buf)'.")

    def __len__(self) -> int:
        return self._len

    def feed(self, d: bytes) -> None:
        if d:
            self._lst.append(d)
            self._len += len(d)

    def _chop(self, i: int, e: int) -> bytes:
        lst = self._lst
        d = lst[i]

        o = b''.join([
            *lst[:i],
            d[:e],
        ])

        self._lst = [
            *([d[e:]] if e < len(d) else []),
            *lst[i + 1:],
        ]

        self._len -= len(o)

        return o

    def read(self, n: ta.Optional[int] = None) -> ta.Optional[bytes]:
        if n is None:
            if not self._lst:
                return b''

            o = b''.join(self._lst)
            self._lst = []
            self._len = 0
            return o

        if not (lst := self._lst):
            return None

        c = 0
        for i, d in enumerate(lst):
            r = n - c
            if (l := len(d)) >= r:
                return self._chop(i, r)
            c += l

        return None

    def read_exact(self, sz: int) -> bytes:
        d = self.read(sz)
        if d is None or len(d) != sz:
            raise EOFError(f'ReadableListBuffer got {"no" if d is None else len(d)}, expected {sz}')
        return d

    def read_until_(self, delim: bytes = b'\n', start_buffer: int = 0) -> ta.Union[bytes, int]:
        if not (lst := self._lst):
            return 0

        i = start_buffer
        while i < len(lst):
            if (p := lst[i].find(delim)) >= 0:
                return self._chop(i, p + len(delim))
            i += 1

        return i

    def read_until(self, delim: bytes = b'\n') -> ta.Optional[bytes]:
        r = self.read_until_(delim)
        return r if isinstance(r, bytes) else None

    #

    DEFAULT_BUFFERED_READER_CHUNK_SIZE: ta.ClassVar[int] = -1

    @ta.final
    class _BufferedBytesReader(BufferedBytesReader):
        def __init__(
                self,
                raw: RawBytesReader,
                buf: 'ReadableListBuffer',
                *,
                chunk_size: ta.Optional[int] = None,
        ) -> None:
            self._raw = raw
            self._buf = buf
            self._chunk_size = chunk_size or ReadableListBuffer.DEFAULT_BUFFERED_READER_CHUNK_SIZE

        def read1(self, n: int = -1, /) -> bytes:
            if n < 0:
                n = self._chunk_size
            if not n:
                return b''
            if 0 < n <= len(self._buf):
                return self._buf.read(n) or b''
            return self._raw.read1(n)

        def read(self, /, n: int = -1) -> bytes:
            if n < 0:
                return self.readall()
            while len(self._buf) < n:
                if not (b := self._raw.read1(n)):
                    break
                self._buf.feed(b)

            if len(self._buf) >= n:
                return self._buf.read(n) or b''

            # EOF with a partial buffer: return what we have.
            return self._buf.read() or b''

        def readall(self) -> bytes:
            buf = io.BytesIO()
            buf.write(self._buf.read() or b'')
            while (b := self._raw.read1(self._chunk_size)):
                buf.write(b)
            return buf.getvalue()

    def new_buffered_reader(
            self,
            raw: RawBytesReader,
            *,
            chunk_size: ta.Optional[int] = None,
    ) -> BufferedBytesReader:
        return self._BufferedBytesReader(
            raw,
            self,
            chunk_size=chunk_size,
        )

    @ta.final
    class _AsyncBufferedBytesReader(AsyncBufferedBytesReader):
        def __init__(
                self,
                raw: AsyncRawBytesReader,
                buf: 'ReadableListBuffer',
                *,
                chunk_size: ta.Optional[int] = None,
        ) -> None:
            self._raw = raw
            self._buf = buf
            self._chunk_size = chunk_size or ReadableListBuffer.DEFAULT_BUFFERED_READER_CHUNK_SIZE

        async def read1(self, n: int = -1, /) -> bytes:
            if n < 0:
                n = self._chunk_size
            if not n:
                return b''
            if 0 < n <= len(self._buf):
                return self._buf.read(n) or b''
            return await self._raw.read1(n)

        async def read(self, /, n: int = -1) -> bytes:
            if n < 0:
                return await self.readall()
            while len(self._buf) < n:
                if not (b := await self._raw.read1(n)):
                    break
                self._buf.feed(b)

            if len(self._buf) >= n:
                return self._buf.read(n) or b''

            # EOF with a partial buffer: return what we have.
            return self._buf.read() or b''

        async def readall(self) -> bytes:
            buf = io.BytesIO()
            buf.write(self._buf.read() or b'')
            while b := await self._raw.read1(self._chunk_size):
                buf.write(b)
            return buf.getvalue()

    def new_async_buffered_reader(
            self,
            raw: AsyncRawBytesReader,
            *,
            chunk_size: ta.Optional[int] = None,
    ) -> AsyncBufferedBytesReader:
        return self._AsyncBufferedBytesReader(
            raw,
            self,
            chunk_size=chunk_size,
        )


##


class IncrementalWriteBuffer:
    def __init__(
            self,
            data: bytes,
            *,
            write_size: int = 0x10000,
    ) -> None:
        super().__init__()

        check.not_empty(data)
        self._len = len(data)
        self._write_size = write_size

        self._lst = [
            data[i:i + write_size]
            for i in range(0, len(data), write_size)
        ]
        self._pos = 0

    @property
    def rem(self) -> int:
        return self._len - self._pos

    def write(self, fn: ta.Callable[[bytes], int]) -> int:
        lst = check.not_empty(self._lst)

        t = 0
        for i, d in enumerate(lst):  # noqa
            d = check.not_empty(d)
            n = fn(d)
            if not n:
                break

            if n > len(d):
                raise ValueError(n)

            t += n

            if n < len(d):
                # Short write - keep the remainder of this chunk and stop.
                self._lst = [
                    d[n:],
                    *lst[i + 1:],
                ]
                self._pos += t
                return t

        if t:
            # Only fully-written chunks were consumed.
            self._lst = lst[i + 1:]
            self._pos += t

        return t



import abc
import typing as ta

from ... import lang
from ...metadata import ObjectMetadata
from ...metadata import append_object_metadata
from ...metadata import get_object_metadata
from .. import asgi


T = ta.TypeVar('T')


##


class AppMarker(ObjectMetadata, lang.Abstract):
    pass


##


def append_app_marker(obj: T, *markers: AppMarker) -> T:
    append_object_metadata(obj, *markers)
    return obj


def get_app_markers(obj: ta.Any) -> ta.Sequence[AppMarker]:
    return get_object_metadata(obj, type=AppMarker)


##


class AppMarkerProcessor(lang.Abstract):
    @abc.abstractmethod
    def process_app(self, app: asgi.App) -> asgi.App:
        raise NotImplementedError


class NopAppMarkerProcessor(AppMarkerProcessor, lang.Final):
    def process_app(self, app: asgi.App) -> asgi.App:
        return app


##


AppMarkerProcessorMap: ta.TypeAlias = ta.Mapping[type[AppMarker], AppMarkerProcessor]



# ruff: noqa: UP045
# @omlish-lite
import threading
import typing as ta

from ..events import AsyncliteEvent
from ..events import AsyncliteEvents
from .base import SyncAsyncliteApi
from .base import SyncAsyncliteObject


##


class SyncAsyncliteEvent(AsyncliteEvent, SyncAsyncliteObject):
    def __init__(self, u: threading.Event) -> None:
        super().__init__()

        self._u = u

    def set(self) -> None:
        return self._u.set()

    def is_set(self) -> bool:
        return self._u.is_set()

    async def wait(self, *, timeout: ta.Optional[float] = None) -> None:
        if not self._u.wait(timeout=timeout):
            raise TimeoutError


class SyncAsyncliteEvents(AsyncliteEvents, SyncAsyncliteApi):
    def make_event(self) -> AsyncliteEvent:
        return SyncAsyncliteEvent(threading.Event())



import abc
import contextlib
import typing as ta

from .. import check
from .. import collections as col
from .. import lang
from ..logs import all as logs
from .debug import _ResourcesDebug


T = ta.TypeVar('T')
U = ta.TypeVar('U')

BaseResourceManagerT = ta.TypeVar('BaseResourceManagerT', bound='BaseResourceManager')


log = logs.get_module_logger(globals())


##


class ResourceManagerRef(lang.Abstract):
    pass


class ResourceManagerRefNotRegisteredError(Exception):
    pass


##


class BaseResourceManager(
    _ResourcesDebug,
    lang.Abstract,
    lang.NotPicklable,
    ta.Generic[T],
):
    """Essentially a reference-tracked [Async]ContextManager."""

    def __init__(
            self,
            *,
            init_ref: ResourceManagerRef | None = None,
            no_autoclose: bool = False,
    ) -> None:
        super().__init__()

        self._no_autoclose = no_autoclose

        self._closed = False

        self._refs: ta.MutableSet[ResourceManagerRef] = col.IdentitySet()

        if init_ref is not None:
            self.add_ref(init_ref)

    @abc.abstractmethod
    def init(self) -> T:
        raise NotImplementedError

    @property
    def autoclose(self) -> bool:
        return not self._no_autoclose

    @property
    def num_refs(self) -> int:
        return len(self._refs)

    @property
    def closed(self) -> bool:
        return self._closed

    def __repr__(self) -> str:
        return lang.attr_repr(self, 'closed', 'num_refs', with_id=True)

    #

    class _InitRef(ResourceManagerRef):
        pass

    #

    def add_ref(self, ref: ResourceManagerRef) -> None:
        check.isinstance(ref, ResourceManagerRef)
        check.state(not self._closed)

        self._refs.add(ref)

    def has_ref(self, ref: ResourceManagerRef) -> bool:
        return ref in self._refs

    @abc.abstractmethod
    def remove_ref(self, ref: ResourceManagerRef) -> T:
        raise NotImplementedError

    #

    @abc.abstractmethod
    def enter_context(self, cm: ta.ContextManager[U]) -> U:
        raise NotImplementedError

    #

    @abc.abstractmethod
    def new_managed(self, v: U) -> 'BaseResourceManaged[U, ta.Self]':
        raise NotImplementedError

    #

    def __del__(self) -> None:
        if not self._closed:
            ref_lst = list(self._refs)
            log.error(
                f'{__package__}.{self.__class__.__name__}.__del__: '  # noqa
                f'%r deleted without being closed! '
                f'refs: %s',
                repr(self),
                ref_lst,
            )


class BaseResourceManaged(
    ResourceManagerRef,
    _ResourcesDebug,
    lang.Abstract,
    lang.NotPicklable,
    ta.Generic[U, BaseResourceManagerT],
):
    """
    A class to 'handoff' a ref to a `BaseResourceManager`, allowing the `BaseResourceManager` to temporarily survive
    being passed from instantiation within a callee.

    This class wraps an arbitrary value, likely an object referencing resources managed by the `BaseResourceManager`,
    which is accessed by `__enter__/__aenter__`'ing. However, as the point of this class is handoff of a
    `BaseResourceManager`, not necessarily some arbitrary value, the value needn't necessarily be related to the
    `BaseResourceManager`, or may even be `None`.

    The ref to the `BaseResourceManager` is allocated in the ctor, so the contract is that an instance of this must be
    immediately `__enter__/__aenter__`'d before doing anything else with the return value of the call. Failure to do so
    leaks the `BaseResourceManager`.
    """

    def __init__(self, v: U, resources: BaseResourceManagerT) -> None:
        super().__init__()

        self._v = v
        self._resources = resources

        resources.add_ref(self)

    _state: ta.Literal['new', 'entered', 'exited'] = 'new'

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}<{self._v!r}, {self._state}>'

    def __del__(self) -> None:
        if self._state != 'exited':
            log.error(
                f'{__package__}.{self.__class__.__name__}.__del__: '  # noqa
                f'%r deleted without being entered and exited! '
                f'resources: %s',
                repr(self),
                repr(self._resources),
            )


##


@ta.final
class ResourceManager(
    BaseResourceManager[None],
    lang.Final,
):
    def __init__(
            self,
            *,
            init_ref: ResourceManagerRef | None = None,
            no_autoclose: bool = False,
    ) -> None:
        super().__init__(
            init_ref=init_ref,
            no_autoclose=no_autoclose,
        )

        self._es = contextlib.ExitStack()

    def init(self) -> None:
        self._es.__enter__()

        self._init_debug()

    #

    @classmethod
    def new(cls, **kwargs: ta.Any) -> ta.ContextManager['ResourceManager']:
        @contextlib.contextmanager
        def inner():
            init_ref = BaseResourceManager._InitRef()  # noqa

            res = ResourceManager(init_ref=init_ref, **kwargs)

            res.init()

            try:
                yield res

            finally:
                res.remove_ref(init_ref)

        return inner()

    @classmethod
    def or_new(
            cls,
            resources: ta.Optional['ResourceManager'],
            **kwargs: ta.Any,
    ) -> ta.ContextManager['ResourceManager']:
        if resources is None:
            return cls.new(**kwargs)

        @contextlib.contextmanager
        def inner():
            yield resources

        return inner()

    #

    def remove_ref(self, ref: ResourceManagerRef) -> None:
        check.isinstance(ref, ResourceManagerRef)

        try:
            self._refs.remove(ref)

        except KeyError:
            raise ResourceManagerRefNotRegisteredError(ref) from None

        if not self._no_autoclose and not self._refs:
            self.close()

    #

    def enter_context(self, cm: ta.ContextManager[U]) -> U:
        check.state(not self._closed)

        return self._es.enter_context(cm)

    #

    def new_managed(self, v: U) -> 'ResourceManaged[U]':
        return ResourceManaged(v, self)  # noqa

    #

    def close(self) -> None:
        try:
            self._es.__exit__(None, None, None)
        finally:
            self._closed = True


@ta.final
class ResourceManaged(
    BaseResourceManaged[U, ResourceManager],
    lang.Final,
    ta.ContextManager[U],
):
    def __enter__(self) -> U:
        check.state(self._state == 'new')
        self._state = 'entered'

        self._init_debug()

        return self._v

    def __exit__(self, exc_type, exc_val, exc_tb):
        check.state(self._state == 'entered')
        self._state = 'exited'

        self._resources.remove_ref(self)


##


@ta.final
class AsyncResourceManager(
    BaseResourceManager[ta.Awaitable[None]],
    lang.Final,
):
    def __init__(
            self,
            *,
            init_ref: ResourceManagerRef | None = None,
            no_autoclose: bool = False,
    ) -> None:
        super().__init__(
            init_ref=init_ref,
            no_autoclose=no_autoclose,
        )

        self._aes = contextlib.AsyncExitStack()

    async def init(self) -> None:
        await self._aes.__aenter__()

        self._init_debug()

    #

    @classmethod
    def new(cls, **kwargs: ta.Any) -> ta.AsyncContextManager['AsyncResourceManager']:
        @contextlib.asynccontextmanager
        async def inner():
            init_ref = BaseResourceManager._InitRef()  # noqa

            res = AsyncResourceManager(init_ref=init_ref, **kwargs)

            await res.init()

            try:
                yield res

            finally:
                await res.remove_ref(init_ref)

        return inner()

    @classmethod
    def or_new(
            cls,
            resources: ta.Optional['AsyncResourceManager'],
            **kwargs: ta.Any,
    ) -> ta.AsyncContextManager['AsyncResourceManager']:
        if resources is None:
            return cls.new(**kwargs)

        @contextlib.asynccontextmanager
        async def inner():
            yield resources

        return inner()

    #

    async def remove_ref(self, ref: ResourceManagerRef) -> None:
        check.isinstance(ref, ResourceManagerRef)

        try:
            self._refs.remove(ref)

        except KeyError:
            raise ResourceManagerRefNotRegisteredError(ref) from None

        if not self._no_autoclose and not self._refs:
            await self.aclose()

    #

    def enter_context(self, cm: ta.ContextManager[U]) -> U:
        check.state(not self._closed)

        return self._aes.enter_context(cm)

    async def enter_async_context(self, cm: ta.AsyncContextManager[U]) -> U:
        check.state(not self._closed)

        return await self._aes.enter_async_context(cm)

    #

    def new_managed(self, v: U) -> 'AsyncResourceManaged[U]':
        return AsyncResourceManaged(v, self)  # noqa

    #

    async def aclose(self) -> None:
        try:
            await self._aes.__aexit__(None, None, None)
        finally:
            self._closed = True


@ta.final
class AsyncResourceManaged(
    BaseResourceManaged[U, AsyncResourceManager],
    lang.Final,
    ta.AsyncContextManager[U],
):
    async def __aenter__(self) -> U:
        check.state(self._state == 'new')
        self._state = 'entered'

        self._init_debug()

        return self._v

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        check.state(self._state == 'entered')
        self._state = 'exited'

        await self._resources.remove_ref(self)



# ruff: noqa: UP006 UP007 UP043 UP045
# @omlish-lite
##
# MIT License
#
# Copyright (c) 2019 Masaaki Goshima
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
##
import dataclasses as dc
import enum
import typing as ta

from ....lite.check import check
from .errors import EofYamlError
from .errors import YamlError
from .errors import YamlErrorOr
from .errors import yaml_error
from .tokens import YAML_RESERVED_TAG_KEYWORD_MAP
from .tokens import YamlIndicator
from .tokens import YamlPosition
from .tokens import YamlToken
from .tokens import YamlTokenMakers
from .tokens import YamlTokens
from .tokens import YamlTokenType
from .tokens import yaml_new_token


##


@dc.dataclass()
class InvalidTokenYamlError(YamlError):
    token: YamlToken

    @property
    def message(self) -> str:
        return check.not_none(self.token.error).message


def _yaml_err_invalid_token(tk: YamlToken) -> InvalidTokenYamlError:
    return InvalidTokenYamlError(
        token=tk,
    )


##


# Context at scanning
@dc.dataclass()
class YamlScanningContext:
    idx: int = 0
    size: int = 0
    not_space_char_pos: int = 0
    not_space_org_char_pos: int = 0
    src: ta.List[str] = dc.field(default_factory=list)
    buf: ta.List[str] = dc.field(default_factory=list)
    obuf: ta.List[str] = dc.field(default_factory=list)
    tokens: YamlTokens = dc.field(default_factory=YamlTokens)
    mstate: ta.Optional['YamlMultiLineState'] = None

    def clear(self) -> None:
        self.reset_buffer()
        self.mstate = None

    def reset(self, src: ta.List[str]) -> None:
        self.idx = 0
        self.size = len(src)
        self.src = list(src)
        self.tokens = YamlTokens()
        self.reset_buffer()
        self.mstate = None

    def reset_buffer(self) -> None:
        self.buf = []
        self.obuf = []
        self.not_space_char_pos = 0
        self.not_space_org_char_pos = 0

    def break_multi_line(self) -> None:
        self.mstate = None

    def get_multi_line_state(self) -> ta.Optional['YamlMultiLineState']:
        return self.mstate

    def set_literal(self, last_delim_column: int, opt: str) -> None:
        mstate = YamlMultiLineState(
            is_literal=True,
            opt=opt,
        )
        indent = _yaml_first_line_indent_column_by_opt(opt)
        if indent > 0:
            mstate.first_line_indent_column = last_delim_column + indent
        self.mstate = mstate

    def set_folded(self, last_delim_column: int, opt: str) -> None:
        mstate = YamlMultiLineState(
            is_folded=True,
            opt=opt,
        )
        indent = _yaml_first_line_indent_column_by_opt(opt)
        if indent > 0:
            mstate.first_line_indent_column = last_delim_column + indent
        self.mstate = mstate

    def set_raw_folded(self, column: int) -> None:
        mstate = YamlMultiLineState(
            is_raw_folded=True,
        )
        mstate.update_indent_column(column)
        self.mstate = mstate

    def add_token(self, tk: ta.Optional[YamlToken]) -> None:
        if tk is None:
            return
        self.tokens.append(tk)  # FIXME: .add??

    def add_buf(self, r: str) -> None:
        if len(self.buf) == 0 and (r == ' ' or r == '\t'):
            return
        self.buf += r
        if r != ' ' and r != '\t':
            self.not_space_char_pos = len(self.buf)

    def add_buf_with_tab(self, r: str) -> None:
        if len(self.buf) == 0 and r == ' ':
            return
        self.buf += r
        if r != ' ':
            self.not_space_char_pos = len(self.buf)

    def add_origin_buf(self, r: str) -> None:
        self.obuf += r
        if r != ' ' and r != '\t':
            self.not_space_org_char_pos = len(self.obuf)

    def remove_right_space_from_buf(self) -> None:
        trimmed_buf = self.obuf[:self.not_space_org_char_pos]
        buflen = len(trimmed_buf)
        diff = len(self.obuf) - buflen
        if diff > 0:
            self.obuf = self.obuf[:buflen]
            self.buf = list(self.buffered_src())

    def is_eos(self) -> bool:
        return len(self.src) - 1 <= self.idx

    def is_next_eos(self) -> bool:
        return len(self.src) <= self.idx + 1

    def next(self) -> bool:
        return self.idx < self.size

    def source(self, s: int, e: int) -> str:
        return ''.join(self.src[s:e])

    def previous_char(self) -> str:
        if self.idx > 0:
            return self.src[self.idx - 1]
        return ''

    def current_char(self) -> str:
        if self.size > self.idx:
            return self.src[self.idx]
        return ''

    def next_char(self) -> str:
        if self.size > self.idx + 1:
            return self.src[self.idx + 1]
        return ''

    def repeat_num(self, r: str) -> int:
        cnt = 0
        for i in range(self.idx, self.size):
            if self.src[i] == r:
                cnt += 1
            else:
                break
        return cnt

    def progress(self, num: int) -> None:
        self.idx += num

    def exists_buffer(self) -> bool:
        return len(self.buffered_src()) != 0

    def is_multi_line(self) -> bool:
        return self.mstate is not None

    def buffered_src(self) -> ta.List[str]:
        src = self.buf[:self.not_space_char_pos]

        if self.is_multi_line():
            mstate = check.not_none(self.get_multi_line_state())

            # remove end '\n' character and trailing empty lines.
            # https://yaml.org/spec/1.2.2/#8112-block-chomping-indicator
            if mstate.has_trim_all_end_newline_opt():
                # If the '-' flag is specified, all trailing newline characters will be removed.
                src = list(''.join(src).rstrip('\n'))

            elif not mstate.has_keep_all_end_newline_opt():
                # Normally, all but one of the trailing newline characters are removed.
                new_line_char_count = 0
                for i in range(len(src) - 1, -1, -1):
                    if src[i] == '\n':
                        new_line_char_count += 1
                        continue
                    break

                removed_new_line_char_count = new_line_char_count - 1
                while removed_new_line_char_count > 0:
                    src = list(''.join(src).rstrip('\n'))
                    removed_new_line_char_count -= 1

            # If the text ends with a space character, remove all of them.
            if mstate.has_trim_all_end_newline_opt():
                src = list(''.join(src).rstrip(' '))

            if src == ['\n']:
                # If the content consists only of a newline, it can be considered as the document ending without any
                # specified value, so it is treated as an empty string.
                src = []

            if mstate.has_keep_all_end_newline_opt() and len(src) == 0:
                src = ['\n']

        return src

    def buffered_token(self, pos: YamlPosition) -> ta.Optional[YamlToken]:
        if self.idx == 0:
            return None

        source = self.buffered_src()
        if len(source) == 0:
            self.buf = self.buf[:0]  # clear value's buffer only.
            return None

        tk: ta.Optional[YamlToken]
        if self.is_multi_line():
            tk = YamlTokenMakers.new_string(''.join(source), ''.join(self.obuf), pos)
        else:
            tk = yaml_new_token(''.join(source), ''.join(self.obuf), pos)

        self.set_token_type_by_prev_tag(tk)
        self.reset_buffer()
        return tk

    def set_token_type_by_prev_tag(self, tk: ta.Optional[YamlToken]) -> None:
        last_tk = self.last_token()
        if last_tk is None:
            return

        if last_tk.type != YamlTokenType.TAG:
            return

        tag = last_tk.value
        if tag not in YAML_RESERVED_TAG_KEYWORD_MAP:
            check.not_none(tk).type = YamlTokenType.STRING

    def last_token(self) -> ta.Optional[YamlToken]:
        if len(self.tokens) != 0:
            return self.tokens[-1]

        return None

    @staticmethod
    def new(src: ta.List[str]) -> 'YamlScanningContext':
        ctx = YamlScanningContext()
        ctx.reset(src)
        return ctx


##


@dc.dataclass()
class YamlMultiLineState:
    opt: str = ''
    first_line_indent_column: int = 0
    prev_line_indent_column: int = 0
    line_indent_column: int = 0
    last_not_space_only_line_indent_column: int = 0
    space_only_indent_column: int = 0
    folded_new_line: bool = False
    is_raw_folded: bool = False
    is_literal: bool = False
    is_folded: bool = False

    def last_delim_column(self) -> int:
        if self.first_line_indent_column == 0:
            return 0
        return self.first_line_indent_column - 1

    def update_indent_column(self, column: int) -> None:
        if self.first_line_indent_column == 0:
            self.first_line_indent_column = column
        if self.line_indent_column == 0:
            self.line_indent_column = column

    def update_space_only_indent_column(self, column: int) -> None:
        if self.first_line_indent_column != 0:
            return
        self.space_only_indent_column = column

    def validate_indent_after_space_only(self, column: int) -> ta.Optional[YamlError]:
        if self.first_line_indent_column != 0:
            return None
        if self.space_only_indent_column > column:
            return yaml_error('invalid number of indent is specified after space only')
        return None

    def validate_indent_column(self) -> ta.Optional[YamlError]:
        if _yaml_first_line_indent_column_by_opt(self.opt) == 0:
            return None
        if self.first_line_indent_column > self.line_indent_column:
            return yaml_error('invalid number of indent is specified in the multi-line header')
        return None

    def update_new_line_state(self) -> None:
        self.prev_line_indent_column = self.line_indent_column
        if self.line_indent_column != 0:
            self.last_not_space_only_line_indent_column = self.line_indent_column
        self.folded_new_line = True
        self.line_indent_column = 0

    def is_indent_column(self, column: int) -> bool:
        if self.first_line_indent_column == 0:
            return column == 1
        return self.first_line_indent_column > column

    def add_indent(self, ctx: YamlScanningContext, column: int) -> None:
        if self.first_line_indent_column == 0:
            return

        # If the first line of the document has already been evaluated, the number is treated as the threshold, since
        # the `first_line_indent_column` is a positive number.
        if column < self.first_line_indent_column:
            return

        # `c.folded_new_line` is a variable that is set to True for every newline.
        if not self.is_literal and self.folded_new_line:
            self.folded_new_line = False

        # Since add_buf ignore space character, add to the buffer directly.
        ctx.buf += ' '
        ctx.not_space_char_pos = len(ctx.buf)

    # update_new_line_in_folded if Folded or RawFolded context and the content on the current line starts at the same
    # column as the previous line, treat the new-line-char as a space.
    def update_new_line_in_folded(self, ctx: YamlScanningContext, column: int) -> None:
        if self.is_literal:
            return

        # Folded or RawFolded.

        if not self.folded_new_line:
            return

        last_char = ''
        prev_last_char = ''
        if len(ctx.buf) != 0:
            last_char = ctx.buf[-1]
        if len(ctx.buf) > 1:
            prev_last_char = ctx.buf[-2]

        if self.line_indent_column == self.prev_line_indent_column:
            # ---
            # >
            #  a
            #  b
            if last_char == '\n':
                ctx.buf[-1] = ' '

        elif self.prev_line_indent_column == 0 and self.last_not_space_only_line_indent_column == column:
            # if previous line is indent-space and new-line-char only, prev_line_indent_column is zero. In this case,
            # last new-line-char is removed.
            # ---
            # >
            #  a
            #
            #  b
            if last_char == '\n' and prev_last_char == '\n':
                ctx.buf = ctx.buf[:len(ctx.buf) - 1]
                ctx.not_space_char_pos = len(ctx.buf)

        self.folded_new_line = False

    def has_trim_all_end_newline_opt(self) -> bool:
        return self.opt.startswith('-') or self.opt.endswith('-') or self.is_raw_folded

    def has_keep_all_end_newline_opt(self) -> bool:
        return self.opt.startswith('+') or self.opt.endswith('+')


##


def _yaml_first_line_indent_column_by_opt(opt: str) -> int:
    opt = opt.lstrip('-')
    opt = opt.lstrip('+')
    opt = opt.rstrip('-')
    opt = opt.rstrip('+')
    try:
        return int(opt, 10)
    except ValueError:
        return 0


##


class YamlIndentState(enum.Enum):
    # EQUAL equals previous indent
    EQUAL = enum.auto()
    # UP more indent than previous
    UP = enum.auto()
    # DOWN less indent than previous
    DOWN = enum.auto()
    # KEEP uses not indent token
    KEEP = enum.auto()


# Scanner holds the scanner's internal state while processing a given text. It can be allocated as part of another data
# structure but must be initialized via init before use.
@dc.dataclass()
class YamlScanner:
    source: ta.List[str] = dc.field(default_factory=list)
    source_pos: int = 0
    source_size: int = 0
    # line number. This number starts from 1.
    line: int = 0
    # column number. This number starts from 1.
    column: int = 0
    # offset represents the offset from the beginning of the source.
    offset: int = 0
    # last_delim_column is the last column needed to compare indent is retained.
    last_delim_column: int = 0
    # indent_num indicates the number of spaces used for indentation.
    indent_num: int = 0
    # prev_line_indent_num indicates the number of spaces used for indentation at previous line.
    prev_line_indent_num: int = 0
    # indent_level indicates the level of indent depth. This value does not match the column value.
    indent_level: int = 0
    is_first_char_at_line: bool = False
    is_anchor: bool = False
    is_alias: bool = False
    is_directive: bool = False
    started_flow_sequence_num: int = 0
    started_flow_map_num: int = 0
    indent_state: YamlIndentState = YamlIndentState.EQUAL
    saved_pos: ta.Optional[YamlPosition] = None

    def pos(self) -> YamlPosition:
        return YamlPosition(
            line=self.line,
            column=self.column,
            offset=self.offset,
            indent_num=self.indent_num,
            indent_level=self.indent_level,
        )

    def buffered_token(self, ctx: YamlScanningContext) -> ta.Optional[YamlToken]:
        if self.saved_pos is not None:
            tk = ctx.buffered_token(self.saved_pos)
            self.saved_pos = None
            return tk

        line = self.line
        column = self.column - len(ctx.buf)
        level = self.indent_level
        if ctx.is_multi_line():
            line -= self.new_line_count(ctx.buf)
            column = ''.join(ctx.obuf).find(''.join(ctx.buf)) + 1
            # Since we are in a literal, folded or raw folded we can use the indent level from the last token.
            last = ctx.last_token()
            if last is not None:  # The last token should never be None here.
                level = last.position.indent_level + 1

        return ctx.buffered_token(YamlPosition(
            line=line,
            column=column,
            offset=self.offset - len(ctx.buf),
            indent_num=self.indent_num,
            indent_level=level,
        ))

    def progress_column(self, ctx: YamlScanningContext, num: int) -> None:
        self.column += num
        self.offset += num
        self.progress(ctx, num)

    def progress_only(self, ctx: YamlScanningContext, num: int) -> None:
        self.offset += num
        self.progress(ctx, num)

    def progress_line(self, ctx: YamlScanningContext) -> None:
        self.prev_line_indent_num = self.indent_num
        self.column = 1
        self.line += 1
        self.offset += 1
        self.indent_num = 0
        self.is_first_char_at_line = True
        self.is_anchor = False
        self.is_alias = False
        self.is_directive = False
        self.progress(ctx, 1)

    def progress(self, ctx: YamlScanningContext, num: int) -> None:
        ctx.progress(num)
        self.source_pos += num

    def is_new_line_char(self, c: str) -> bool:
        if c == '\n':
            return True
        if c == '\r':
            return True
        return False

    def new_line_count(self, src: ta.List[str]) -> int:
        size = len(src)
        cnt = 0
        i = -1
        while True:
            i += 1
            if not (i < size):
                break
            c = src[i]
            if c == '\r':
                if i + 1 < size and src[i + 1] == '\n':
                    i += 1
                cnt += 1
            elif c == '\n':
                cnt += 1
        return cnt

    def update_indent_level(self) -> None:
        if self.prev_line_indent_num < self.indent_num:
            self.indent_level += 1
        elif self.prev_line_indent_num > self.indent_num:
            if self.indent_level > 0:
                self.indent_level -= 1

    def update_indent_state(self, ctx: YamlScanningContext) -> None:
        if self.last_delim_column == 0:
            return

        if self.last_delim_column < self.column:
            self.indent_state = YamlIndentState.UP
        else:
            # If last_delim_column and self.column are the same, treat as Down state since it is the same column as
            # delimiter.
            self.indent_state = YamlIndentState.DOWN

    def update_indent(self, ctx: YamlScanningContext, c: str) -> None:
        if self.is_first_char_at_line and self.is_new_line_char(c):
            return
        if self.is_first_char_at_line and c == ' ':
            self.indent_num += 1
            return
        if self.is_first_char_at_line and c == '\t':
            # Found tab indent. In this case, scan_tab returns error.
            return
        if not self.is_first_char_at_line:
            self.indent_state = YamlIndentState.KEEP
            return
        self.update_indent_level()
        self.update_indent_state(ctx)
        self.is_first_char_at_line = False

    def is_changed_to_indent_state_down(self) -> bool:
        return self.indent_state == YamlIndentState.DOWN

    def is_changed_to_indent_state_up(self) -> bool:
        return self.indent_state == YamlIndentState.UP

    def add_buffered_token_if_exists(self, ctx: YamlScanningContext) -> None:
        ctx.add_token(self.buffered_token(ctx))

    def break_multi_line(self, ctx: YamlScanningContext) -> None:
        ctx.break_multi_line()

    def scan_single_quote(self, ctx: YamlScanningContext) -> YamlErrorOr[YamlToken]:
        ctx.add_origin_buf("'")
        srcpos = self.pos()
        start_index = ctx.idx + 1
        src = ctx.src
        size = len(src)
        value: ta.List[str] = []
        is_first_line_char = False
        is_new_line = False

        idx = start_index - 1
        while True:
            idx += 1
            if not (idx < size):
                break

            if not is_new_line:
                self.progress_column(ctx, 1)
            else:
                is_new_line = False

            c = src[idx]
            ctx.add_origin_buf(c)
            if self.is_new_line_char(c):
                not_space_idx = -1
                for i in range(len(value) - 1, -1, -1):
                    if value[i] == ' ':
                        continue
                    not_space_idx = i
                    break

                if len(value) > not_space_idx:
                    value = value[:not_space_idx + 1]
                if is_first_line_char:
                    value += '\n'
                else:
                    value += ' '

                is_first_line_char = True
                is_new_line = True
                self.progress_line(ctx)
                if idx + 1 < size:
                    if (err := self.validate_document_separator_marker(ctx, src[idx + 1:])) is not None:
                        return err

                continue

            if is_first_line_char and c == ' ':
                continue

            if is_first_line_char and c == '\t':
                if self.last_delim_column >= self.column:
                    return _yaml_err_invalid_token(
                        YamlTokenMakers.new_invalid(
                            yaml_error('tab character cannot be used for indentation in single-quoted text'),
                            ''.join(ctx.obuf),
                            self.pos(),
                        ),
                    )

                continue

            if c != "'":
                value += c
                is_first_line_char = False
                continue

            if idx + 1 < len(ctx.src) and ctx.src[idx + 1] == '\'':
                # '' handle as ' character
                value += c
                ctx.add_origin_buf(c)
                idx += 1
                self.progress_column(ctx, 1)
                continue

            self.progress_column(ctx, 1)
            return YamlTokenMakers.new_single_quote(''.join(value), ''.join(ctx.obuf), srcpos)

        self.progress_column(ctx, 1)
        return _yaml_err_invalid_token(
            YamlTokenMakers.new_invalid(
                yaml_error('could not find end character of single-quoted text'),
                ''.join(ctx.obuf),
                srcpos,
            ),
        )

    def scan_double_quote(self, ctx: YamlScanningContext) -> YamlErrorOr[YamlToken]:
        ctx.add_origin_buf('"')
        srcpos = self.pos()
        start_index = ctx.idx + 1
        src = ctx.src
        size = len(src)
        value: ta.List[str] = []
        is_first_line_char = False
        is_new_line = False

        idx = start_index - 1
        while True:
            idx += 1
            if not (idx < size):
                break

            if not is_new_line:
                self.progress_column(ctx, 1)
            else:
                is_new_line = False

            c = src[idx]
            ctx.add_origin_buf(c)
            if self.is_new_line_char(c):
                not_space_idx = -1
                for i in range(len(value) - 1, -1, -1):
                    if value[i] == ' ':
                        continue
                    not_space_idx = i
                    break

                if len(value) > not_space_idx:
                    value = value[:not_space_idx + 1]

                if is_first_line_char:
                    value += '\n'
                else:
                    value += ' '

                is_first_line_char = True
                is_new_line = True
                self.progress_line(ctx)
                if idx + 1 < size:
                    if (err := self.validate_document_separator_marker(ctx, src[idx + 1:])) is not None:
                        return err

                continue

            if is_first_line_char and c == ' ':
                continue

            if is_first_line_char and c == '\t':
                if self.last_delim_column >= self.column:
                    return _yaml_err_invalid_token(
                        YamlTokenMakers.new_invalid(
                            yaml_error('tab character cannot be used for indentation in double-quoted text'),
                            ''.join(ctx.obuf),
                            self.pos(),
                        ),
                    )

                continue

            if c == '\\':
                is_first_line_char = False
                if idx + 1 >= size:
                    value += c
                    continue

                next_char = src[idx + 1]
                progress = 0

                if next_char == '0':
                    progress = 1
                    ctx.add_origin_buf(next_char)
                    value += chr(0)
                elif next_char == 'a':
                    progress = 1
                    ctx.add_origin_buf(next_char)
                    value += '\x07'
                elif next_char == 'b':
                    progress = 1
                    ctx.add_origin_buf(next_char)
                    value += '\x08'
                elif next_char == 't':
                    progress = 1
                    ctx.add_origin_buf(next_char)
                    value += '\x09'
                elif next_char == 'n':
                    progress = 1
                    ctx.add_origin_buf(next_char)
                    value += '\x0A'
                elif next_char == 'v':
                    progress = 1
                    ctx.add_origin_buf(next_char)
                    value += '\x0B'
                elif next_char == 'f':
                    progress = 1
                    ctx.add_origin_buf(next_char)
                    value += '\x0C'
                elif next_char == 'r':
                    progress = 1
                    ctx.add_origin_buf(next_char)
                    value += '\x0D'
                elif next_char == 'e':
                    progress = 1
                    ctx.add_origin_buf(next_char)
                    value += '\x1B'
                elif next_char == ' ':
                    progress = 1
                    ctx.add_origin_buf(next_char)
                    value += '\x20'
                elif next_char == '"':
                    progress = 1
                    ctx.add_origin_buf(next_char)
                    value += '\x22'
                elif next_char == '/':
                    progress = 1
                    ctx.add_origin_buf(next_char)
                    value += '\x2F'
                elif next_char == '\\':
                    progress = 1
                    ctx.add_origin_buf(next_char)
                    value += '\x5C'
                elif next_char == 'N':
                    progress = 1
                    ctx.add_origin_buf(next_char)
                    value += '\x85'
                elif next_char == '_':
                    progress = 1
                    ctx.add_origin_buf(next_char)
                    value += '\xA0'
                elif next_char == 'L':
                    progress = 1
                    ctx.add_origin_buf(next_char)
                    value += '\u2028'
                elif next_char == 'P':
                    progress = 1
                    ctx.add_origin_buf(next_char)
                    value += '\u2029'

                elif next_char == 'x':
                    if idx + 3 >= size:
                        progress = 1
                        ctx.add_origin_buf(next_char)
                        value += next_char
                    else:
                        progress = 3
                        code_num = _yaml_hex_runes_to_int(src[idx + 2: idx + progress + 1])
                        value += chr(code_num)

                elif next_char == 'u':
                    # \u0000 style must have 5 characters at least.
                    if idx + 5 >= size:
                        return _yaml_err_invalid_token(
                            YamlTokenMakers.new_invalid(
                                yaml_error('not enough length for escaped UTF-16 character'),
                                ''.join(ctx.obuf),
                                self.pos(),
                            ),
                        )

                    progress = 5
                    code_num = _yaml_hex_runes_to_int(src[idx + 2: idx + 6])

                    # handle surrogate pairs.
                    if code_num >= 0xD800 and code_num <= 0xDBFF:
                        high = code_num

                        # \u0000\u0000 style must have 11 characters at least.
                        if idx + 11 >= size:
                            return _yaml_err_invalid_token(
                                YamlTokenMakers.new_invalid(
                                    yaml_error('not enough length for escaped UTF-16 surrogate pair'),
                                    ''.join(ctx.obuf),
                                    self.pos(),
                                ),
                            )

                        if src[idx + 6] != '\\' or src[idx + 7] != 'u':
                            return _yaml_err_invalid_token(
                                YamlTokenMakers.new_invalid(
                                    yaml_error('found unexpected character after high surrogate for UTF-16 surrogate pair'),  # noqa
                                    ''.join(ctx.obuf),
                                    self.pos(),
                                ),
                            )

                        low = _yaml_hex_runes_to_int(src[idx + 8: idx + 12])
                        if low < 0xDC00 or low > 0xDFFF:
                            return _yaml_err_invalid_token(
                                YamlTokenMakers.new_invalid(
                                    yaml_error('found unexpected low surrogate after high surrogate'),
                                    ''.join(ctx.obuf),
                                    self.pos(),
                                ),
                            )

                        code_num = ((high - 0xD800) * 0x400) + (low - 0xDC00) + 0x10000
                        progress += 6

                    value += chr(code_num)

                elif next_char == 'U':
                    # \U00000000 style must have 9 characters at least.
                    if idx + 9 >= size:
                        return _yaml_err_invalid_token(
                            YamlTokenMakers.new_invalid(
                                yaml_error('not enough length for escaped UTF-32 character'),
                                ''.join(ctx.obuf),
                                self.pos(),
                            ),
                        )

                    progress = 9
                    code_num = _yaml_hex_runes_to_int(src[idx + 2: idx + 10])
                    value += chr(code_num)

                elif next_char == '\n':
                    is_first_line_char = True
                    is_new_line = True
                    ctx.add_origin_buf(next_char)
                    self.progress_column(ctx, 1)
                    self.progress_line(ctx)
                    idx += 1
                    continue

                elif next_char == '\r':
                    is_first_line_char = True
                    is_new_line = True
                    ctx.add_origin_buf(next_char)
                    self.progress_line(ctx)
                    progress = 1
                    # Skip \n after \r in CRLF sequences
                    if idx + 2 < size and src[idx + 2] == '\n':
                        ctx.add_origin_buf('\n')
                        progress = 2

                elif next_char == '\t':
                    progress = 1
                    ctx.add_origin_buf(next_char)
                    value += next_char

                else:
                    self.progress_column(ctx, 1)
                    return _yaml_err_invalid_token(
                        YamlTokenMakers.new_invalid(
                            yaml_error(f'found unknown escape character {next_char!r}'),
                            ''.join(ctx.obuf),
                            self.pos(),
                        ),
                    )

                idx += progress
                self.progress_column(ctx, progress)
                continue

            if c == '\t':
                found_not_space_char = False
                progress = 0

                for i in range(idx + 1, size):
                    if src[i] == ' ' or src[i] == '\t':
                        progress += 1
                        continue

                    if self.is_new_line_char(src[i]):
                        break

                    found_not_space_char = True

                if found_not_space_char:
                    value += c
                    if src[idx + 1] != '"':
                        self.progress_column(ctx, 1)

                else:
                    idx += progress
                    self.progress_column(ctx, progress)

                continue

            if c != '"':
                value += c
                is_first_line_char = False
                continue

            self.progress_column(ctx, 1)
            return YamlTokenMakers.new_double_quote(''.join(value), ''.join(ctx.obuf), srcpos)

        self.progress_column(ctx, 1)
        return _yaml_err_invalid_token(
            YamlTokenMakers.new_invalid(
                yaml_error('could not find end character of double-quoted text'),
                ''.join(ctx.obuf),
                srcpos,
            ),
        )

    def validate_document_separator_marker(self, ctx: YamlScanningContext, src: ta.List[str]) -> ta.Optional[YamlError]:
        if self.found_document_separator_marker(src):
            return _yaml_err_invalid_token(
                YamlTokenMakers.new_invalid(
                    yaml_error('found unexpected document separator'),
                    ''.join(ctx.obuf),
                    self.pos(),
                ),
            )

        return None

    def found_document_separator_marker(self, src: ta.List[str]) -> bool:
        if len(src) < 3:
            return False

        marker = ''
        if len(src) == 3:
            marker = ''.join(src)
        else:
            marker = _yaml_trim_right_func(''.join(src[:4]), lambda r: r == ' ' or r == '\t' or r == '\n' or r == '\r')

        return marker == '---' or marker == '...'

    def scan_quote(self, ctx: YamlScanningContext, ch: str) -> YamlErrorOr[bool]:
        if ctx.exists_buffer():
            return False

        if ch == "'":
            tk = self.scan_single_quote(ctx)
            if isinstance(tk, YamlError):
                return tk

            ctx.add_token(tk)

        else:
            tk = self.scan_double_quote(ctx)
            if isinstance(tk, YamlError):
                return tk

            ctx.add_token(tk)

        ctx.clear()
        return True

    def scan_white_space(self, ctx: YamlScanningContext) -> bool:
        if ctx.is_multi_line():
            return False

        if not self.is_anchor and not self.is_directive and not self.is_alias and not self.is_first_char_at_line:
            return False

        if self.is_first_char_at_line:
            self.progress_column(ctx, 1)
            ctx.add_origin_buf(' ')
            return True

        if self.is_directive:
            self.add_buffered_token_if_exists(ctx)
            self.progress_column(ctx, 1)
            ctx.add_origin_buf(' ')
            return True

        self.add_buffered_token_if_exists(ctx)
        self.is_anchor = False
        self.is_alias = False
        return True

    def is_merge_key(self, ctx: YamlScanningContext) -> bool:
        if ctx.repeat_num('<') != 2:
            return False

        src = ctx.src
        size = len(src)
        for idx in range(ctx.idx + 2, size):
            c = src[idx]
            if c == ' ':
                continue

            if c != ':':
                return False

            if idx + 1 < size:
                nc = src[idx + 1]
                if nc == ' ' or self.is_new_line_char(nc):
                    return True

        return False

    def scan_tag(self, ctx: YamlScanningContext) -> YamlErrorOr[bool]:
        if ctx.exists_buffer() or self.is_directive:
            return False

        ctx.add_origin_buf('!')
        self.progress(ctx, 1)  # skip '!' character

        progress = 0
        for idx, c in enumerate(ctx.src[ctx.idx:]):
            progress = idx + 1

            if c == ' ':
                ctx.add_origin_buf(c)
                value = ctx.source(ctx.idx - 1, ctx.idx + idx)
                ctx.add_token(YamlTokenMakers.new_tag(value, ''.join(ctx.obuf), self.pos()))
                self.progress_column(ctx, len(value))
                ctx.clear()
                return True

            elif c == ',':
                if self.started_flow_sequence_num > 0 or self.started_flow_map_num > 0:
                    value = ctx.source(ctx.idx - 1, ctx.idx + idx)
                    ctx.add_token(YamlTokenMakers.new_tag(value, ''.join(ctx.obuf), self.pos()))
                    # progress column before collect-entry for scanning it at scan_flow_entry function.
                    self.progress_column(ctx, len(value) - 1)
                    ctx.clear()
                    return True
                else:
                    ctx.add_origin_buf(c)

            elif c in ('\n', '\r'):
                ctx.add_origin_buf(c)
                value = ctx.source(ctx.idx - 1, ctx.idx + idx)
                ctx.add_token(YamlTokenMakers.new_tag(value, ''.join(ctx.obuf), self.pos()))
                # progress column before new-line-char for scanning new-line-char at scan_new_line function.
                self.progress_column(ctx, len(value) - 1)
                ctx.clear()
                return True

            elif c in ('{', '}'):
                ctx.add_origin_buf(c)
                self.progress_column(ctx, progress)
                invalid_tk = YamlTokenMakers.new_invalid(
                    yaml_error(f'found invalid tag character {c!r}'),
                    ''.join(ctx.obuf),
                    self.pos(),
                )
                return _yaml_err_invalid_token(invalid_tk)

            else:
                ctx.add_origin_buf(c)

        self.progress_column(ctx, progress)
        ctx.clear()
        return True

    def scan_comment(self, ctx: YamlScanningContext) -> bool:
        if ctx.exists_buffer():
            c = ctx.previous_char()
            if c != ' ' and c != '\t' and not self.is_new_line_char(c):
                return False

        self.add_buffered_token_if_exists(ctx)
        ctx.add_origin_buf('#')
        self.progress(ctx, 1)  # skip '#' character

        for idx, c in enumerate(ctx.src[ctx.idx:]):
            ctx.add_origin_buf(c)
            if not self.is_new_line_char(c):
                continue
            if ctx.previous_char() == '\\':
                continue

            value = ctx.source(ctx.idx, ctx.idx + idx)
            progress = len(value)
            ctx.add_token(YamlTokenMakers.new_comment(''.join(value), ''.join(ctx.obuf), self.pos()))
            self.progress_column(ctx, progress)
            self.progress_line(ctx)
            ctx.clear()
            return True

        # document ends with comment.
        value = ''.join(ctx.src[ctx.idx:])
        ctx.add_token(YamlTokenMakers.new_comment(value, ''.join(ctx.obuf), self.pos()))
        progress = len(value)
        self.progress_column(ctx, progress)
        self.progress_line(ctx)
        ctx.clear()
        return True

    def scan_multi_line(self, ctx: YamlScanningContext, c: str) -> ta.Optional[YamlError]:
        state = check.not_none(ctx.get_multi_line_state())
        ctx.add_origin_buf(c)

        # normalize CR and CRLF to LF
        if c == '\r':
            if ctx.next_char() == '\n':
                ctx.add_origin_buf('\n')
                self.progress(ctx, 1)
                self.offset += 1

            c = '\n'

        if ctx.is_eos():
            if self.is_first_char_at_line and c == ' ':
                state.add_indent(ctx, self.column)
            else:
                ctx.add_buf(c)

            state.update_indent_column(self.column)
            if (err := state.validate_indent_column()) is not None:
                invalid_tk = YamlTokenMakers.new_invalid(yaml_error(str(err)), ''.join(ctx.obuf), self.pos())
                self.progress_column(ctx, 1)
                return _yaml_err_invalid_token(invalid_tk)

            value = ctx.buffered_src()
            ctx.add_token(YamlTokenMakers.new_string(''.join(value), ''.join(ctx.obuf), self.pos()))
            ctx.clear()
            self.progress_column(ctx, 1)

        elif self.is_new_line_char(c):
            ctx.add_buf(c)
            state.update_space_only_indent_column(self.column - 1)
            state.update_new_line_state()
            self.progress_line(ctx)
            if ctx.next():
                if self.found_document_separator_marker(ctx.src[ctx.idx:]):
                    value = ctx.buffered_src()
                    ctx.add_token(YamlTokenMakers.new_string(''.join(value), ''.join(ctx.obuf), self.pos()))
                    ctx.clear()
                    self.break_multi_line(ctx)

        elif self.is_first_char_at_line and c == ' ':
            state.add_indent(ctx, self.column)
            self.progress_column(ctx, 1)

        elif self.is_first_char_at_line and c == '\t' and state.is_indent_column(self.column):
            err = _yaml_err_invalid_token(
                YamlTokenMakers.new_invalid(
                    yaml_error('found a tab character where an indentation space is expected'),
                    ''.join(ctx.obuf),
                    self.pos(),
                ),
            )
            self.progress_column(ctx, 1)
            return err

        elif c == '\t' and not state.is_indent_column(self.column):
            ctx.add_buf_with_tab(c)
            self.progress_column(ctx, 1)

        else:
            if (err := state.validate_indent_after_space_only(self.column)) is not None:
                invalid_tk = YamlTokenMakers.new_invalid(yaml_error(str(err)), ''.join(ctx.obuf), self.pos())
                self.progress_column(ctx, 1)
                return _yaml_err_invalid_token(invalid_tk)

            state.update_indent_column(self.column)
            if (err := state.validate_indent_column()) is not None:
                invalid_tk = YamlTokenMakers.new_invalid(yaml_error(str(err)), ''.join(ctx.obuf), self.pos())
                self.progress_column(ctx, 1)
                return _yaml_err_invalid_token(invalid_tk)

            if (col := state.last_delim_column()) > 0:
                self.last_delim_column = col

            state.update_new_line_in_folded(ctx, self.column)
            ctx.add_buf_with_tab(c)
            self.progress_column(ctx, 1)

        return None

    def scan_new_line(self, ctx: YamlScanningContext, c: str) -> None:
        if len(ctx.buf) > 0 and self.saved_pos is None:
            buf_len = len(ctx.buffered_src())
            self.saved_pos = self.pos()
            self.saved_pos.column -= buf_len
            self.saved_pos.offset -= buf_len

        # if the following case, origin buffer has unnecessary two spaces.
        # So, `removeRightSpaceFromOriginBuf` remove them, also fix column number too.
        # ---
        # a:[space][space]
        #   b: c
        ctx.remove_right_space_from_buf()

        # There is no problem that we ignore CR which followed by LF and normalize it to LF, because of following
        # YAML1.2 spec.
        # > Line breaks inside scalar content must be normalized by the YAML processor. Each such line break must be
        #   parsed into a single line feed character.
        # > Outside scalar content, YAML allows any line break to be used to terminate lines.
        # > -- https://yaml.org/spec/1.2/spec.html
        if c == '\r' and ctx.next_char() == '\n':
            ctx.add_origin_buf('\r')
            self.progress(ctx, 1)
            self.offset += 1
            c = '\n'

        if ctx.is_eos():
            self.add_buffered_token_if_exists(ctx)
        elif self.is_anchor or self.is_alias or self.is_directive:
            self.add_buffered_token_if_exists(ctx)

        if ctx.exists_buffer() and self.is_first_char_at_line:
            if ctx.buf[-1] == ' ':
                ctx.buf[-1] = '\n'
            else:
                ctx.buf += '\n'
        else:
            ctx.add_buf(' ')

        ctx.add_origin_buf(c)
        self.progress_line(ctx)

    def is_flow_mode(self) -> bool:
        if self.started_flow_sequence_num > 0:
            return True

        if self.started_flow_map_num > 0:
            return True

        return False

    def scan_flow_map_start(self, ctx: YamlScanningContext) -> bool:
        if ctx.exists_buffer() and not self.is_flow_mode():
            return False

        self.add_buffered_token_if_exists(ctx)
        ctx.add_origin_buf('{')
        ctx.add_token(YamlTokenMakers.new_mapping_start(''.join(ctx.obuf), self.pos()))
        self.started_flow_map_num += 1
        self.progress_column(ctx, 1)
        ctx.clear()
        return True

    def scan_flow_map_end(self, ctx: YamlScanningContext) -> bool:
        if self.started_flow_map_num <= 0:
            return False

        self.add_buffered_token_if_exists(ctx)
        ctx.add_origin_buf('}')
        ctx.add_token(YamlTokenMakers.new_mapping_end(''.join(ctx.obuf), self.pos()))
        self.started_flow_map_num -= 1
        self.progress_column(ctx, 1)
        ctx.clear()
        return True

    def scan_flow_array_start(self, ctx: YamlScanningContext) -> bool:
        if ctx.exists_buffer() and not self.is_flow_mode():
            return False

        self.add_buffered_token_if_exists(ctx)
        ctx.add_origin_buf('[')
        ctx.add_token(YamlTokenMakers.new_sequence_start(''.join(ctx.obuf), self.pos()))
        self.started_flow_sequence_num += 1
        self.progress_column(ctx, 1)
        ctx.clear()
        return True

    def scan_flow_array_end(self, ctx: YamlScanningContext) -> bool:
        if ctx.exists_buffer() and self.started_flow_sequence_num <= 0:
            return False

        self.add_buffered_token_if_exists(ctx)
        ctx.add_origin_buf(']')
        ctx.add_token(YamlTokenMakers.new_sequence_end(''.join(ctx.obuf), self.pos()))
        self.started_flow_sequence_num -= 1
        self.progress_column(ctx, 1)
        ctx.clear()
        return True

    def scan_flow_entry(self, ctx: YamlScanningContext, c: str) -> bool:
        if self.started_flow_sequence_num <= 0 and self.started_flow_map_num <= 0:
            return False

        self.add_buffered_token_if_exists(ctx)
        ctx.add_origin_buf(c)
        ctx.add_token(YamlTokenMakers.new_collect_entry(''.join(ctx.obuf), self.pos()))
        self.progress_column(ctx, 1)
        ctx.clear()
        return True

    def scan_map_delim(self, ctx: YamlScanningContext) -> YamlErrorOr[bool]:
        nc = ctx.next_char()
        if self.is_directive or self.is_anchor or self.is_alias:
            return False

        if (
                self.started_flow_map_num <= 0 and
                nc != ' ' and
                nc != '\t' and
                not self.is_new_line_char(nc) and
                not ctx.is_next_eos()
        ):
            return False

        if self.started_flow_map_num > 0 and nc == '/':
            # like http://
            return False

        if self.started_flow_map_num > 0:
            tk = ctx.last_token()
            if tk is not None and tk.type == YamlTokenType.MAPPING_VALUE:
                return False

        if ''.join(ctx.obuf).lstrip(' ').startswith('\t') and not ''.join(ctx.buf).startswith('\t'):
            invalid_tk = YamlTokenMakers.new_invalid(
                yaml_error('tab character cannot use as a map key directly'),
                ''.join(ctx.obuf),
                self.pos(),
            )
            self.progress_column(ctx, 1)
            return _yaml_err_invalid_token(invalid_tk)

        # mapping value
        tk = self.buffered_token(ctx)
        if tk is not None:
            self.last_delim_column = tk.position.column
            ctx.add_token(tk)

        elif (tk := ctx.last_token()) is not None:
            # If the map key is quote, the buffer does not exist because it has already been cut into tokens.
            # Therefore, we need to check the last token.
            if tk.indicator == YamlIndicator.QUOTED_SCALAR:
                self.last_delim_column = tk.position.column

        ctx.add_token(YamlTokenMakers.new_mapping_value(self.pos()))
        self.progress_column(ctx, 1)
        ctx.clear()
        return True

    def scan_document_start(self, ctx: YamlScanningContext) -> bool:
        if self.indent_num != 0:
            return False

        if self.column != 1:
            return False

        if ctx.repeat_num('-') != 3:
            return False

        if ctx.size > ctx.idx + 3:
            c = ctx.src[ctx.idx + 3]
            if c != ' ' and c != '\t' and c != '\n' and c != '\r':
                return False

        self.add_buffered_token_if_exists(ctx)
        ctx.add_token(YamlTokenMakers.new_document_header(''.join(ctx.obuf) + '---', self.pos()))
        self.progress_column(ctx, 3)
        ctx.clear()
        self.clear_state()
        return True

    def scan_document_end(self, ctx: YamlScanningContext) -> bool:
        if self.indent_num != 0:
            return False

        if self.column != 1:
            return False

        if ctx.repeat_num('.') != 3:
            return False

        self.add_buffered_token_if_exists(ctx)
        ctx.add_token(YamlTokenMakers.new_document_end(''.join(ctx.obuf) + '...', self.pos()))
        self.progress_column(ctx, 3)
        ctx.clear()
        return True

    def scan_merge_key(self, ctx: YamlScanningContext) -> bool:
        if not self.is_merge_key(ctx):
            return False

        self.last_delim_column = self.column
        ctx.add_token(YamlTokenMakers.new_merge_key(''.join(ctx.obuf) + '<<', self.pos()))
        self.progress_column(ctx, 2)
        ctx.clear()
        return True

    def scan_raw_folded_char(self, ctx: YamlScanningContext) -> bool:
        if not ctx.exists_buffer():
            return False

        if not self.is_changed_to_indent_state_up():
            return False

        ctx.set_raw_folded(self.column)
        ctx.add_buf('-')
        ctx.add_origin_buf('-')
        self.progress_column(ctx, 1)
        return True

    def scan_sequence(self, ctx: YamlScanningContext) -> YamlErrorOr[bool]:
        if ctx.exists_buffer():
            return False

        nc = ctx.next_char()
        if nc != 0 and nc != ' ' and nc != '\t' and not self.is_new_line_char(nc):
            return False

        if ''.join(ctx.obuf).lstrip(' ').startswith('\t'):
            invalid_tk = YamlTokenMakers.new_invalid(
                yaml_error('tab character cannot use as a sequence delimiter'),
                ''.join(ctx.obuf),
                self.pos(),
            )
            self.progress_column(ctx, 1)
            return _yaml_err_invalid_token(invalid_tk)

        self.add_buffered_token_if_exists(ctx)
        ctx.add_origin_buf('-')
        tk = YamlTokenMakers.new_sequence_entry(''.join(ctx.obuf), self.pos())
        self.last_delim_column = tk.position.column
        ctx.add_token(tk)
        self.progress_column(ctx, 1)
        ctx.clear()
        return True

    def scan_multi_line_header(self, ctx: YamlScanningContext) -> YamlErrorOr[bool]:
        if ctx.exists_buffer():
            return False

        if (err := self.scan_multi_line_header_option(ctx)) is not None:
            return err

        self.progress_line(ctx)
        return True

    def validate_multi_line_header_option(self, opt: str) -> ta.Optional[YamlError]:
        if len(opt) == 0:
            return None

        org_opt = opt
        opt = opt.lstrip('-')
        opt = opt.lstrip('+')
        opt = opt.rstrip('-')
        opt = opt.rstrip('+')
        if len(opt) == 0:
            return None

        if opt == '0':
            return yaml_error(f'invalid header option: {org_opt!r}')

        try:
            i = int(opt, 10)
        except ValueError:
            return yaml_error(f'invalid header option: {org_opt!r}')

        if i > 9:
            return yaml_error(f'invalid header option: {org_opt!r}')

        return None

    def scan_multi_line_header_option(self, ctx: YamlScanningContext) -> ta.Optional[YamlError]:
        header = ctx.current_char()
        ctx.add_origin_buf(header)
        self.progress(ctx, 1)  # skip '|' or '>' character

        progress = 0
        crlf = False
        for idx, c in enumerate(ctx.src[ctx.idx:]):
            progress = idx
            ctx.add_origin_buf(c)
            if self.is_new_line_char(c):
                next_idx = ctx.idx + idx + 1
                if c == '\r' and next_idx < len(ctx.src) and ctx.src[next_idx] == '\n':
                    crlf = True
                    continue  # process \n in the next iteration

                break

        end_pos = ctx.idx + progress
        if crlf:
            # Exclude \r
            end_pos = end_pos - 1

        value = ctx.source(ctx.idx, end_pos).rstrip(' ')
        comment_value_index = value.find('#')
        opt = value
        if comment_value_index > 0:
            opt = value[:comment_value_index]

        opt = _yaml_trim_right_func(opt, lambda r: r == ' ' or r == '\t')

        if len(opt) != 0:
            if (err := self.validate_multi_line_header_option(opt)) is not None:
                invalid_tk = YamlTokenMakers.new_invalid(yaml_error(str(err)), ''.join(ctx.obuf), self.pos())
                self.progress_column(ctx, progress)
                return _yaml_err_invalid_token(invalid_tk)

        if self.column == 1:
            self.last_delim_column = 1

        try:
            comment_index = ctx.obuf.index('#')
        except ValueError:
            comment_index = -1
        header_buf = ''.join(ctx.obuf)
        if comment_index > 0:
            header_buf = header_buf[:comment_index]

        if header == '|':
            ctx.add_token(YamlTokenMakers.new_literal('|' + opt, header_buf, self.pos()))
            ctx.set_literal(self.last_delim_column, opt)
        elif header == '>':
            ctx.add_token(YamlTokenMakers.new_folded('>' + opt, header_buf, self.pos()))
            ctx.set_folded(self.last_delim_column, opt)

        if comment_index > 0:
            comment = value[comment_value_index + 1:]
            self.offset += len(header_buf)
            self.column += len(header_buf)
            ctx.add_token(YamlTokenMakers.new_comment(comment, ''.join(ctx.obuf[len(header_buf):]), self.pos()))

        self.indent_state = YamlIndentState.KEEP
        ctx.reset_buffer()
        self.progress_column(ctx, progress)
        return None

    def scan_map_key(self, ctx: YamlScanningContext) -> bool:
        if ctx.exists_buffer():
            return False

        nc = ctx.next_char()
        if nc != ' ' and nc != '\t':
            return False

        tk = YamlTokenMakers.new_mapping_key(self.pos())
        self.last_delim_column = tk.position.column
        ctx.add_token(tk)
        self.progress_column(ctx, 1)
        ctx.clear()
        return True

    def scan_directive(self, ctx: YamlScanningContext) -> bool:
        if ctx.exists_buffer():
            return False
        if self.indent_num != 0:
            return False

        self.add_buffered_token_if_exists(ctx)
        ctx.add_origin_buf('%')
        ctx.add_token(YamlTokenMakers.new_directive(''.join(ctx.obuf), self.pos()))
        self.progress_column(ctx, 1)
        ctx.clear()
        self.is_directive = True
        return True

    def scan_anchor(self, ctx: YamlScanningContext) -> bool:
        if ctx.exists_buffer():
            return False

        self.add_buffered_token_if_exists(ctx)
        ctx.add_origin_buf('&')
        ctx.add_token(YamlTokenMakers.new_anchor(''.join(ctx.obuf), self.pos()))
        self.progress_column(ctx, 1)
        self.is_anchor = True
        ctx.clear()
        return True

    def scan_alias(self, ctx: YamlScanningContext) -> bool:
        if ctx.exists_buffer():
            return False

        self.add_buffered_token_if_exists(ctx)
        ctx.add_origin_buf('*')
        ctx.add_token(YamlTokenMakers.new_alias(''.join(ctx.obuf), self.pos()))
        self.progress_column(ctx, 1)
        self.is_alias = True
        ctx.clear()
        return True

    def scan_reserved_char(self, ctx: YamlScanningContext, c: str) -> ta.Optional[YamlError]:
        if ctx.exists_buffer():
            return None

        ctx.add_buf(c)
        ctx.add_origin_buf(c)
        err = _yaml_err_invalid_token(
            YamlTokenMakers.new_invalid(
                yaml_error(f'{c!r} is a reserved character'),
                ''.join(ctx.obuf),
                self.pos(),
            ),
        )
        self.progress_column(ctx, 1)
        ctx.clear()
        return err

    def scan_tab(self, ctx: YamlScanningContext, c: str) -> ta.Optional[YamlError]:
        if self.started_flow_sequence_num > 0 or self.started_flow_map_num > 0:
            # tabs character is allowed in flow mode.
            return None

        if not self.is_first_char_at_line:
            return None

        ctx.add_buf(c)
        ctx.add_origin_buf(c)
        err = _yaml_err_invalid_token(
            YamlTokenMakers.new_invalid(
                yaml_error("found character '\t' that cannot start any token"),
                ''.join(ctx.obuf),
                self.pos(),
            ),
        )
        self.progress_column(ctx, 1)
        ctx.clear()
        return err

    def _scan(self, ctx: YamlScanningContext) -> ta.Optional[YamlError]:
        while ctx.next():
            c = ctx.current_char()
            # First, change the IndentState.
            # If the target character is the first character in a line, IndentState is Up/Down/Equal state.
            # The second and subsequent letters are Keep.
            self.update_indent(ctx, c)

            # If IndentState is down, tokens are split, so the buffer accumulated until that point needs to be cutted as
            # a token.
            if self.is_changed_to_indent_state_down():
                self.add_buffered_token_if_exists(ctx)

            if ctx.is_multi_line():
                if self.is_changed_to_indent_state_down():
                    if (tk := ctx.last_token()) is not None:
                        # If literal/folded content is empty, no string token is added.
                        # Therefore, add an empty string token.
                        # But if literal/folded token column is 1, it is invalid at down state.
                        if tk.position.column == 1:
                            return yaml_error(_yaml_err_invalid_token(
                                YamlTokenMakers.new_invalid(
                                    yaml_error('could not find multi-line content'),
                                    ''.join(ctx.obuf),
                                    self.pos(),
                                ),
                            ))

                        if tk.type != YamlTokenType.STRING:
                            ctx.add_token(YamlTokenMakers.new_string('', '', self.pos()))

                    self.break_multi_line(ctx)

                else:
                    if (err := self.scan_multi_line(ctx, c)) is not None:
                        return err

                    continue

            if c == '{':
                if self.scan_flow_map_start(ctx):
                    continue

            elif c == '}':
                if self.scan_flow_map_end(ctx):
                    continue

            elif c == '.':
                if self.scan_document_end(ctx):
                    continue

            elif c == '<':
                if self.scan_merge_key(ctx):
                    continue

            elif c == '-':
                if self.scan_document_start(ctx):
                    continue

                if self.scan_raw_folded_char(ctx):
                    continue

                scanned = self.scan_sequence(ctx)
                if isinstance(scanned, YamlError):
                    return scanned

                if scanned:
                    continue

            elif c == '[':
                if self.scan_flow_array_start(ctx):
                    continue

            elif c == ']':
                if self.scan_flow_array_end(ctx):
                    continue

            elif c == ',':
                if self.scan_flow_entry(ctx, c):
                    continue

            elif c == ':':
                scanned = self.scan_map_delim(ctx)
                if isinstance(scanned, YamlError):
                    return scanned

                if scanned:
                    continue

            elif c in ('|', '>'):
                scanned = self.scan_multi_line_header(ctx)
                if isinstance(scanned, YamlError):
                    return scanned

                if scanned:
                    continue

            elif c == '!':
                scanned = self.scan_tag(ctx)
                if isinstance(scanned, YamlError):
                    return scanned

                if scanned:
                    continue

            elif c == '%':
                if self.scan_directive(ctx):
                    continue

            elif c == '?':
                if self.scan_map_key(ctx):
                    continue

            elif c == '&':
                if self.scan_anchor(ctx):
                    continue

            elif c == '*':
                if self.scan_alias(ctx):
                    continue

            elif c == '#':
                if self.scan_comment(ctx):
                    continue

            elif c in ("'", '"'):
                scanned = self.scan_quote(ctx, c)
                if isinstance(scanned, YamlError):
                    return scanned

                if scanned:
                    continue

            elif c in ('\r', '\n'):
                self.scan_new_line(ctx, c)
                continue

            elif c == ' ':
                if self.scan_white_space(ctx):
                    continue

            elif c in ('@', '`'):
                if (err := self.scan_reserved_char(ctx, c)) is not None:
                    return err

            elif c == '\t':
                if ctx.exists_buffer() and self.last_delim_column == 0:
                    # tab indent for plain text (yaml-test-suite's spec-example-7-12-plain-lines).
                    self.indent_num += 1
                    ctx.add_origin_buf(c)
                    self.progress_only(ctx, 1)
                    continue

                if self.last_delim_column < self.column:
                    self.indent_num += 1
                    ctx.add_origin_buf(c)
                    self.progress_only(ctx, 1)
                    continue

                if (err := self.scan_tab(ctx, c)) is not None:
                    return err

            ctx.add_buf(c)
            ctx.add_origin_buf(c)
            self.progress_column(ctx, 1)

        self.add_buffered_token_if_exists(ctx)
        return None

    # init prepares the scanner s to tokenize the text src by setting the scanner at the beginning of src.
    def init(self, text: str) -> None:
        src = text
        self.source = list(src)
        self.source_pos = 0
        self.source_size = len(src)
        self.line = 1
        self.column = 1
        self.offset = 1
        self.is_first_char_at_line = True
        self.clear_state()

    def clear_state(self) -> None:
        self.prev_line_indent_num = 0
        self.last_delim_column = 0
        self.indent_level = 0
        self.indent_num = 0

    # scan scans the next token and returns the token collection. The source end is indicated by io.EOF.
    def scan(self) -> ta.Tuple[ta.Optional[YamlTokens], ta.Optional[YamlError]]:
        if self.source_pos >= self.source_size:
            return None, EofYamlError()

        ctx = YamlScanningContext.new(self.source[self.source_pos:])

        lst = YamlTokens()
        err = self._scan(ctx)
        lst.extend(ctx.tokens)

        if err is not None:
            # var invalidTokenErr *InvalidTokenError
            # if errors.As(err, &invalidTokenErr):
            #     lst = append(lst, invalidTokenErr.Token)
            return lst, err

        return lst, None


# Tokenize split to token instances from string
def yaml_tokenize(src: str) -> YamlTokens:
    s = YamlScanner()
    s.init(src)

    tks = YamlTokens()
    while True:
        sub_tokens, err = s.scan()
        if isinstance(err, EofYamlError):
            break

        tks.add(*check.not_none(sub_tokens))

    return tks


##


def _yaml_hex_to_int(s: str) -> int:
    if len(s) != 1:
        raise ValueError(s)
    b = s[0]
    if 'A' <= b <= 'F':
        return ord(b) - ord('A') + 10
    if 'a' <= b <= 'f':
        return ord(b) - ord('a') + 10
    return ord(b) - ord('0')


def _yaml_hex_runes_to_int(b: ta.List[str]) -> int:
    n = 0
    for i in range(len(b)):
        n += _yaml_hex_to_int(b[i]) << ((len(b) - i - 1) * 4)
    return n


def _yaml_trim_right_func(s: str, predicate: ta.Callable[[str], bool]) -> str:
    if not s:
        return s

    i = len(s) - 1
    while i >= 0 and predicate(s[i]):
        i -= 1

    return s[:i + 1]



# ruff: noqa: UP007 UP045
# @omlish-lite
import abc
import typing as ta

from ...lite.abstract import Abstract


BytesLike = ta.Union[bytes, bytearray, memoryview]  # ta.TypeAlias


##


class ByteStreamBufferLike(Abstract):
    @ta.final
    def __bool__(self) -> bool:
        raise TypeError('Do not use bool() for ByteStreamBufferLike, use len().')

    @abc.abstractmethod
    def __len__(self) -> int:
        """
        Return the number of readable bytes.

        This is expected to be O(1). Many drivers and codecs use `len(buf)` in tight loops to decide whether more data
        is needed before attempting to parse a frame.
        """

        raise NotImplementedError

    @abc.abstractmethod
    def peek(self) -> memoryview:
        """
        Return a contiguous, read-only `memoryview` of the first available bytes.

        This is the "next chunk" fast-path: for segmented views, the returned memoryview may represent only the first
        segment (and thus may be shorter than `len(self)`), but it must be non-copying. This is the fast-path for codecs
        that can parse headers from an initial contiguous region.

        The returned view should be treated as ephemeral: callers must assume it may be invalidated by subsequent buffer
        mutations (advance/write/reserve/commit), depending on the implementation.
        """

        raise NotImplementedError

    @abc.abstractmethod
    def segments(self) -> ta.Sequence[memoryview]:
        """
        Return the readable contents as an ordered sequence of non-copying `memoryview` segments.

        This method is required because efficient operations in pure Python typically depend on delegating work to
        CPython's optimized implementations for searching/slicing within contiguous regions. By exposing
        already-contiguous segments, the buffer enables implementations of `find/rfind` and higher-level framing to
        avoid Python-level per-byte iteration.

        The returned segments must:
          - collectively represent exactly the readable bytes, in order
          - be 1-D, byte-oriented views (itemsize 1)
          - be non-copying views of the underlying storage
          - be non-empty - lack of data is represented by returning no segments, not empty segments

        Callers must assume that the returned views may be invalidated by subsequent mutations of the originating
        buffer/view (e.g., advancing, writing, reserving, committing), depending on the implementation's rules.
        """

        raise NotImplementedError


class ByteStreamBufferView(ByteStreamBufferLike, Abstract):
    """
    A read-only, possibly non-contiguous view of bytes.

    This is the result type of operations like `ByteStreamBuffer.split_to()`: it represents a *logical* byte sequence
    without requiring a copy. A `ByteStreamBufferView` is intentionally minimal: it is not a general-purpose container
    API, not a random-access sequence, and not intended for arbitrary indexing/slicing-heavy use.

    `ByteStreamBufferView` exists to make copy boundaries explicit:
      - Use `segments()` / `peek()` to access data without copying.
      - Use `tobytes()` (or `bytes(view)`) to intentionally materialize a contiguous `bytes` object.

    Implementations may be backed by one or many `memoryview` segments; the semantics are defined as if all readable
    bytes were concatenated in order.
    """

    @abc.abstractmethod
    def tobytes(self) -> bytes:
        """
        Materialize this view as a contiguous `bytes` object (copying).

        This is the explicit copy boundary: callers should prefer `peek()` / `segments()` for zero-copy-ish access when
        feasible, and use `tobytes()` only when a contiguous owned `bytes` is required.
        """

        raise NotImplementedError


class ByteStreamBuffer(ByteStreamBufferLike, Abstract):
    """
    An incremental, consumption-oriented byte accumulator intended for protocol parsing.

    A `ByteStreamBuffer` is a *stream buffer*: bytes are appended by a driver/transport and then consumed by codecs via
    peeking, searching, splitting, and advancing-without forcing repeated concatenation or reallocation. It is
    explicitly designed to support segmented storage (to avoid "a huge buffer pinned by a tiny tail") and to enable
    low-copy pipeline-style decoding (Netty/Tokio-inspired).

    What it is for:
      - buffering raw bytes between I/O and protocol codecs,
      - framing (delimiters/length-prefixed) using split/advance,
      - efficient searching over buffered bytes using C-accelerated primitives via `memoryview` segments.

    What it is *not* for:
      - a general-purpose replacement for `bytes`/`bytearray`,
      - a `collections.abc.Sequence` or random-access container abstraction,
      - arbitrary indexing/slicing-heavy workloads (use `bytes`/`bytearray`/`memoryview` directly).

    `ByteStreamBuffer` deliberately exposes `memoryview` at its boundary. This is foundational: it allows both immutable
    (`bytes`) and mutable (`bytearray`) internal storage to be viewed in O(1) without copying. It also avoids relying
    on `io.BytesIO` as a core backing store: while `BytesIO.getbuffer()` can expose a view, exported views pin the
    underlying buffer against resizing, which makes it awkward as a general-purpose buffer substrate.

    Semantics note:
      Many methods describe behavior in terms of the *conceptual concatenation* of readable bytes, even if the buffer
      is physically segmented. This is what "stream-correct" means here: results must be correct regardless of how the
      buffered bytes are chunked internally.
    """

    @abc.abstractmethod
    def advance(self, n: int, /) -> None:
        """
        Consume (discard) exactly `n` readable bytes from the front of the buffer.

        This operation must not copy remaining bytes unnecessarily. For segmented buffers, this typically adjusts a head
        offset and drops exhausted segments.

        Implementations must raise if `n` is negative or greater than `len(self)`.
        """

        raise NotImplementedError

    @abc.abstractmethod
    def split_to(self, n: int, /) -> ByteStreamBufferView:
        """
        Split off and return a read-only view of the first `n` readable bytes, consuming them from this buffer.

        This is the core "low-copy framing" primitive:
          - codecs can `split_to(frame_len)` to obtain a view of an entire frame without copying,
          - then immediately continue parsing subsequent frames from the remaining bytes.

        Implementations should strive for O(1) or amortized O(1) behavior, returning a view that references underlying
        segments rather than materializing a new contiguous `bytes`.

        Implementations must raise if `n` is negative or greater than `len(self)`.
        """

        raise NotImplementedError

    @abc.abstractmethod
    def coalesce(self, n: int, /) -> memoryview:
        """
        Ensure the first `n` readable bytes are available contiguously and return a view of them.

        Semantics:
          - Non-consuming: does not advance.
          - May restructure internal segments (content-preserving) to make the prefix contiguous.
          - Returns a read-only-ish `memoryview` (callers must not mutate readable bytes).

        Copying behavior:
          - If `peek()` already exposes >= n contiguous bytes, this is zero-copy.
          - Otherwise, it copies exactly the first `n` bytes into a new contiguous segment and rewrites the internal
            segment list so that segment[0] contains that prefix.

        Reserve interaction:
          - Disallowed while an outstanding reservation exists, since reserve() hands out a view that must not be
            invalidated by internal reshaping.
        """

        raise NotImplementedError

    @abc.abstractmethod
    def find(self, sub: bytes, start: int = 0, end: ta.Optional[int] = None) -> int:
        """
        Find the first occurrence of `sub` within the readable bytes and return its offset, or -1 if not found.

        This operation is "stream-correct": it must behave as if searching within the conceptual concatenation of all
        readable bytes, even if the buffer is physically segmented. In particular, matches that span segment boundaries
        must be detected.

        `start` and `end` are offsets into the readable region, matching the semantics of `bytes.find()`:
          - `start` defaults to 0 (the beginning of readable bytes),
          - `end` defaults to `len(self)`.

        Rationale for being part of the core interface:
          In pure Python, higher-level codecs cannot efficiently implement correct cross-segment searching byte-by-byte.
          Keeping `find` near the owning storage allows implementations to exploit contiguous segments and CPython's
          optimized search within each segment while still providing correct stream semantics.
        """

        raise NotImplementedError

    @abc.abstractmethod
    def rfind(self, sub: bytes, start: int = 0, end: ta.Optional[int] = None) -> int:
        """
        Find the last occurrence of `sub` within the readable bytes and return its offset, or -1 if not found.

        This operation is also stream-correct and matches `bytes.rfind()` semantics for `start`/`end`, interpreted as
        offsets into the readable region of this buffer.
        """

        raise NotImplementedError


class MutableByteStreamBuffer(ByteStreamBuffer, Abstract):
    """
    A writable `ByteStreamBuffer`: supports appending bytes and (optionally) reserving writable space.

    `MutableByteStreamBuffer` is the primary target for drivers/transports feeding data into protocol pipelines, and for
    encoders building outbound byte sequences. It intentionally does not imply any particular I/O model (blocking,
    asyncio, custom reactors); it is simply the mutable byte substrate.

    Implementations may be linear (single `bytearray` + indices), segmented (multiple chunks), or adaptive.
    """

    @property
    @abc.abstractmethod
    def max_size(self) -> ta.Optional[int]:
        raise NotImplementedError

    @abc.abstractmethod
    def write(self, data: BytesLike, /) -> None:
        """
        Append `data` to the end of the readable region (after any existing unread bytes).

        Implementations should avoid needless copying; e.g., segmented buffers may store large `bytes` chunks directly,
        while linear buffers may copy into a `bytearray`.
        """

        raise NotImplementedError

    @abc.abstractmethod
    def reserve(self, n: int, /) -> memoryview:
        """
        Reserve writable space for at least `n` bytes and return a writable `memoryview` into that space.

        This method exists to support "close to the metal" drivers that can fill buffers directly (e.g., `recv_into`,
        `readinto`) without allocating temporary `bytes` objects.

        The returned view represents capacity that is not yet part of the readable region. The caller must write into
        some prefix of the view and then call `commit(written)` to make those bytes readable.

        Implementations should document their rules regarding outstanding reservations; a simple and robust rule is:
          - only one active reservation may exist at a time,
          - mutations that would reallocate storage are forbidden while a reservation is outstanding.
        """

        raise NotImplementedError

    @abc.abstractmethod
    def commit(self, n: int, /) -> None:
        """
        Commit `n` bytes from the most recent reservation, making them readable.

        Conceptually, `reserve()` may provide more capacity than the caller actually uses; `commit(n)` "shrinks" that
        over-reservation by only publishing the first `n` bytes as readable.

        Implementations must validate:
          - that a reservation is outstanding,
          - that `0 <= n <= reserved_length`.

        After commit, the reservation is considered consumed; subsequent reads and searches must include the committed
        bytes as part of the readable region.
        """

        raise NotImplementedError



import threading

from .types import ScheduledFn
from .types import ScheduledRunner
from .types import ScheduledRunnerState
from .types import ScheduleHandle


##


class TrackingScheduledRunner(ScheduledRunner):
    def __init__(self, inner: ScheduledRunner) -> None:
        super().__init__()

        self._inner = inner

        self._lock = threading.Lock()
        self._handles: set[ScheduleHandle] = set()

    def get_state(self) -> ScheduledRunnerState:
        return self._inner.get_state()

    def get_handles(self) -> set[ScheduleHandle]:
        with self._lock:
            return set(self._handles)

    def cancel_all_handles(self) -> None:
        with self._lock:
            hs = list(self._handles)

        for h in hs:
            h.cancel()

    def _track_handle(self, h: ScheduleHandle) -> ScheduleHandle:
        with self._lock:
            self._handles.add(h)

        def _done(_h: ScheduleHandle) -> None:
            with self._lock:
                self._handles.discard(_h)

        h.add_done_callback(_done)
        return h

    #

    def schedule(self, delay: float, fn: ScheduledFn) -> ScheduleHandle:
        return self._track_handle(self._inner.schedule(delay, fn))

    def shutdown(
            self,
            *,
            cancel_all: bool = False,
            no_wait: bool = False,
            timeout: float | None = None,
    ) -> None:
        self._inner.shutdown(
            cancel_all=cancel_all,
            no_wait=no_wait,
            timeout=timeout,
        )

    def wait(self, timeout: float | None = None) -> None:
        self._inner.wait(timeout=timeout)



"""
TODO:
 - prebuild field frozenset for getters/setters
  - and one field per line
"""
import dataclasses as dc
import typing as ta
import weakref

from .... import check
from ..._internals import STD_FIELDS_ATTR
from ..._internals import STD_PARAMS_ATTR
from ..generation.base import Generator
from ..generation.base import Plan
from ..generation.base import PlanResult
from ..generation.globals import FROZEN_INSTANCE_ERROR_GLOBAL
from ..generation.idents import CLS_IDENT
from ..generation.idents import IDENT_PREFIX
from ..generation.ops import AddMethodOp
from ..generation.ops import Op
from ..generation.registry import register_generator_type
from ..processing.base import ProcessingContext


##


_UNCHECKED_FROZEN_BASES: ta.MutableSet[type] = weakref.WeakSet()


def unchecked_frozen_base(cls):
    _UNCHECKED_FROZEN_BASES.add(check.isinstance(cls, type))
    return cls


def check_frozen_bases(cls: type, frozen: bool) -> None:
    all_frozen_bases = None
    any_frozen_base = False
    has_dataclass_bases = False

    for b in cls.__mro__[-1:0:-1]:
        if b in _UNCHECKED_FROZEN_BASES:
            continue

        base_fields = getattr(b, STD_FIELDS_ATTR, None)
        if base_fields is None:
            continue

        has_dataclass_bases = True
        if all_frozen_bases is None:
            all_frozen_bases = True

        current_frozen = getattr(b, STD_PARAMS_ATTR).frozen
        all_frozen_bases = all_frozen_bases and current_frozen
        any_frozen_base = any_frozen_base or current_frozen

    if has_dataclass_bases:
        if any_frozen_base and not frozen:
            raise TypeError('cannot inherit non-frozen dataclass from a frozen one')

        if all_frozen_bases is False and frozen:
            raise TypeError('cannot inherit frozen dataclass from a non-frozen one')


##


@dc.dataclass(frozen=True, kw_only=True)
class FrozenPlan(Plan):
    fields: tuple[str, ...]
    allow_dynamic_dunder_attrs: bool


@register_generator_type(FrozenPlan)
class FrozenGenerator(Generator[FrozenPlan]):
    def plan(self, ctx: ProcessingContext) -> PlanResult[FrozenPlan] | None:
        check_frozen_bases(ctx.cls, ctx.cs.frozen)

        if not ctx.cs.frozen:
            return None

        if issubclass(ctx.cls, BaseException):
            raise TypeError('cannot use frozen=True with subclass of BaseException')

        return PlanResult(FrozenPlan(
            fields=tuple(f.name for f in ctx.cs.fields),
            allow_dynamic_dunder_attrs=ctx.cs.allow_dynamic_dunder_attrs,
        ))

    def _generate_one(
            self,
            plan: FrozenPlan,
            mth: str,
            params: ta.Sequence[str],
            exc_args: str,
    ) -> AddMethodOp:
        preamble = []
        condition = []

        # https://github.com/python/cpython/commit/ee6f8413a99d0ee4828e1c81911e203d3fff85d5
        base_condition = f'type(self) is {CLS_IDENT}'

        if plan.allow_dynamic_dunder_attrs:
            condition.extend([
                f'(',
                f'    {base_condition}',
                f'    and not (len(name) > 4 and name[:2] == name[-2:] == "__")',
                f')',
            ])
        else:
            condition.append(base_condition)

        if plan.fields:
            set_ident = f'{IDENT_PREFIX}_{mth}_frozen_fields'
            preamble.extend([
                f'{set_ident} = {{',
                *[
                    f'    {f!r},'
                    for f in plan.fields
                ],
                f'}}',
                f'',
            ])
            condition.append(f'or name in {set_ident}')

        return AddMethodOp(
            f'__{mth}__',
            '\n'.join([
                *preamble,
                f'def __{mth}__(self, {", ".join(params)}):',
                f'    if (',
                *[
                    f'        {l}'
                    for l in condition
                ],
                f'    ):',
                f'        raise {FROZEN_INSTANCE_ERROR_GLOBAL.ident}{exc_args}',
                f'    super({CLS_IDENT}, self).__{mth}__({", ".join(params)})',
            ]),
            frozenset([FROZEN_INSTANCE_ERROR_GLOBAL]),
        )

    def generate(self, plan: FrozenPlan) -> ta.Iterable[Op]:
        return [
            self._generate_one(
                plan,
                'setattr',
                ['name', 'value'],
                '(f"cannot assign to field {name!r}")',
            ),
            self._generate_one(
                plan,
                'delattr',
                ['name'],
                '(f"cannot delete field {name!r}")',
            ),
        ]



import typing as ta

from .. import lang


##


Environ: ta.TypeAlias = ta.Mapping[str, ta.Any]
StartResponse: ta.TypeAlias = ta.Callable[
    [
        str,
        ta.Iterable[tuple[str | bytes, str | bytes]],
        # types.ExceptionInfo | None = None,
    ],
    ta.Callable[[lang.BytesLike], None],
]
App: ta.TypeAlias = ta.Callable[[Environ, StartResponse], ta.Iterable[lang.BytesLike]]


# class App(lang.Abstract):
#     def __enter__(self: AppT) -> AppT:
#         return self
#
#     def __exit__(self, exc_type, exc_val, exc_tb) -> None:
#         return None
#
#     @abc.abstractmethod
#     def __call__(self, environ: Environ, start_response: StartResponse) -> ta.Iterable[bytes]:
#         raise NotImplementedError


# class AsyncApp(lang.Abstract):
#     async def __aenter__(self: AppT) -> AppT:
#         return self
#
#     async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
#         return None
#
#     @abc.abstractmethod
#     async def __call__(self, scope, receive, send) -> ta.Iterable[bytes]:
#         raise NotImplementedError



# @omlish-lite
from ...lite.abstract import Abstract
from .events import AsyncliteEvents
from .identities import AsyncliteIdentities
from .locks import AsyncliteLocks
from .queues import AsyncliteQueues
from .semaphores import AsyncliteSemaphores
from .sleeps import AsyncliteSleeps


##


class Asynclite(
    AsyncliteEvents,
    AsyncliteIdentities,
    AsyncliteLocks,
    AsyncliteQueues,
    AsyncliteSemaphores,
    AsyncliteSleeps,

    Abstract,
):
    pass



"""
https://spec.json5.org/
"""
import io
import json
import typing as ta

from ... import lang
from .errors import Json5Error


##


LITERAL_VALUES: ta.Mapping[str, ta.Any] = {
    'true': True,
    'false': False,
    'null': None,
}


##


STRING_LITERAL_ESCAPES: ta.Mapping[str, str] = {
    'b': '\\u0008',
    'f': '\\u000C',
    'n': '\\u000A',
    'r': '\\u000D',
    't': '\\u0009',
    'v': '\\u000B',
    '0': '\\u0000',
    'u': '\\u',
    '"': '\\"',
    "'": "'",
    '\\': '\\\\',
}


def _check_state(b: bool, fmt: str = 'Json5 error', *args: ta.Any) -> None:
    if not b:
        raise Json5Error(fmt % args)


def translate_string_literal(s: str) -> str:
    _check_state(len(s) > 1)
    q = s[0]
    _check_state(q in '\'"')
    _check_state(s[-1] == q)

    c = 1
    e = len(s) - 1

    b = io.StringIO()
    b.write('"')

    ds = '\\\'"'
    while True:
        n = lang.find_any(s, ds, c, e)
        if n < 0:
            b.write(s[c:e])
            break

        _check_state(n < e)
        b.write(s[c:n])

        x = s[n]
        if x == '\\':
            _check_state(n < (e - 1))

            y = s[n + 1]
            if y in '\n\u2028\u2029':
                c = n + 2

            elif y == '\r':
                c = n + 2
                if c < e and s[c] == '\n':
                    c += 1

            elif y == 'x':
                _check_state(n < (e - 3))
                u = int(s[n + 2:n + 4], 16)
                b.write(f'\\u00{u:02x}')
                c = n + 4

            elif (g := STRING_LITERAL_ESCAPES.get(y)) is not None:
                b.write(g)
                c = n + 2

            elif not ('0' <= y <= '9'):
                b.write(y)
                c = n + 2

            else:
                raise Json5Error(f'Invalid string literal escape: {x}{y}')

        elif x in '\\\'"':
            _check_state(x != q)
            if x == '"':
                b.write('\\"')
            else:
                b.write(x)
            c = n + 1

        else:
            raise RuntimeError

    b.write('"')
    return b.getvalue()


def parse_string_literal(s: str) -> str:
    j = translate_string_literal(s)

    try:
        return json.loads(j)
    except json.JSONDecodeError as e:
        raise Json5Error from e


##


def parse_number_literal(s: str) -> int | float:
    s = s.lower()

    try:
        if 'x' in s:
            return int(s, 16)
        else:
            return float(s)
    except ValueError as e:
        raise Json5Error from e



# Copyright 2010 Pallets
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1.  Redistributions of source code must retain the above copyright notice, this list of conditions and the following
#     disclaimer.
#
# 2.  Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
#     following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3.  Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
#     products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import base64
import dataclasses as dc
import datetime
import decimal
import json as _json
import typing as ta
import uuid

from .. import lang
from .dates import http_date
from .dates import parse_date


if ta.TYPE_CHECKING:
    import markupsafe
else:
    markupsafe = lang.proxy_import('markupsafe')


##


def _default(o: ta.Any) -> ta.Any:
    if isinstance(o, datetime.date):
        return http_date(o)

    if isinstance(o, (decimal.Decimal, uuid.UUID)):
        return str(o)

    if dc.is_dataclass(o):
        return dc.asdict(o)  # type: ignore

    if hasattr(o, '__html__'):
        return str(o.__html__())

    raise TypeError(f'Object of type {type(o).__name__} is not Json serializable')


def json_dumps(obj: ta.Any, **kwargs: ta.Any) -> str:
    kwargs.setdefault('default', _default)
    return _json.dumps(obj, **kwargs)


def json_loads(s: str | bytes, **kwargs: ta.Any) -> ta.Any:
    return _json.loads(s, **kwargs)


##


class JsonTag:
    key: str = ''

    def __init__(self, tagger: 'JsonTagger') -> None:
        super().__init__()

        self.tagger = tagger

    def check(self, value: ta.Any) -> bool:
        raise NotImplementedError

    def to_json(self, value: ta.Any) -> ta.Any:
        raise NotImplementedError

    def to_python(self, value: ta.Any) -> ta.Any:
        raise NotImplementedError

    def tag(self, value: ta.Any) -> dict[str, ta.Any]:
        return {self.key: self.to_json(value)}


class TagDict(JsonTag):
    key = ' di'

    def check(self, value: ta.Any) -> bool:
        return (
            isinstance(value, dict)
            and len(value) == 1
            and next(iter(value)) in self.tagger.tags
        )

    def to_json(self, value: ta.Any) -> ta.Any:
        key = next(iter(value))
        return {f'{key}__': self.tagger.tag(value[key])}

    def to_python(self, value: ta.Any) -> ta.Any:
        key = next(iter(value))
        return {key[:-2]: value[key]}


class PassDict(JsonTag):
    def check(self, value: ta.Any) -> bool:
        return isinstance(value, dict)

    def to_json(self, value: ta.Any) -> ta.Any:
        return {k: self.tagger.tag(v) for k, v in value.items()}

    tag = to_json


class TagTuple(JsonTag):
    key = ' t'

    def check(self, value: ta.Any) -> bool:
        return isinstance(value, tuple)

    def to_json(self, value: ta.Any) -> ta.Any:
        return [self.tagger.tag(item) for item in value]

    def to_python(self, value: ta.Any) -> ta.Any:
        return tuple(value)


class PassList(JsonTag):
    def check(self, value: ta.Any) -> bool:
        return isinstance(value, list)

    def to_json(self, value: ta.Any) -> ta.Any:
        return [self.tagger.tag(item) for item in value]

    tag = to_json


class TagBytes(JsonTag):
    key = ' b'

    def check(self, value: ta.Any) -> bool:
        return isinstance(value, bytes)

    def to_json(self, value: ta.Any) -> ta.Any:
        return base64.b64encode(value).decode('ascii')

    def to_python(self, value: ta.Any) -> ta.Any:
        return base64.b64decode(value)


class TagMarkup(JsonTag):
    key = ' m'

    def check(self, value: ta.Any) -> bool:
        return callable(getattr(value, '__html__', None))

    def to_json(self, value: ta.Any) -> ta.Any:
        return str(value.__html__())

    def to_python(self, value: ta.Any) -> ta.Any:
        return markupsafe.Markup(value)


class TagUuid(JsonTag):
    key = ' u'

    def check(self, value: ta.Any) -> bool:
        return isinstance(value, uuid.UUID)

    def to_json(self, value: ta.Any) -> ta.Any:
        return value.hex

    def to_python(self, value: ta.Any) -> ta.Any:
        return uuid.UUID(value)


class TagDatetime(JsonTag):
    key = ' dt'

    def check(self, value: ta.Any) -> bool:
        return isinstance(value, datetime.datetime)

    def to_json(self, value: ta.Any) -> ta.Any:
        return http_date(value)

    def to_python(self, value: ta.Any) -> ta.Any:
        return parse_date(value)


class JsonTagger:
    default_tags: ta.ClassVar[ta.Sequence[type[JsonTag]]] = [
        TagDict,
        PassDict,
        TagTuple,
        PassList,
        TagBytes,
        *([TagMarkup] if lang.can_import('markupsafe') else []),
        TagUuid,
        TagDatetime,
    ]

    def __init__(self) -> None:
        super().__init__()

        self.tags: dict[str, JsonTag] = {}
        self.order: list[JsonTag] = []

        for cls in self.default_tags:
            self.register(cls)

    def register(
            self,
            tag_class: type[JsonTag],
            force: bool = False,
            index: int | None = None,
    ) -> None:
        tag = tag_class(self)
        key = tag.key

        if key:
            if not force and key in self.tags:
                raise KeyError(f"Tag '{key}' is already registered.")

            self.tags[key] = tag

        if index is None:
            self.order.append(tag)
        else:
            self.order.insert(index, tag)

    def tag(self, value: ta.Any) -> ta.Any:
        for tag in self.order:
            if tag.check(value):
                return tag.tag(value)

        return value

    def untag(self, value: dict[str, ta.Any]) -> ta.Any:
        if len(value) != 1:
            return value

        key = next(iter(value))

        if key not in self.tags:
            return value

        return self.tags[key].to_python(value[key])

    def untag_scan(self, value: ta.Any) -> ta.Any:
        if isinstance(value, dict):
            value = {k: self.untag_scan(v) for k, v in value.items()}
            value = self.untag(value)

        elif isinstance(value, list):
            value = [self.untag_scan(item) for item in value]

        return value

    def dumps(self, value: ta.Any) -> str:
        return json_dumps(self.tag(value), separators=(',', ':'))

    def loads(self, value: str) -> ta.Any:
        return self.untag_scan(json_loads(value))


JSON_TAGGER = JsonTagger()



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
import socket
import typing as ta

from ....io.buffers import IncrementalWriteBuffer
from ....io.buffers import ReadableListBuffer
from ....io.fdio.handlers import SocketFdioHandler
from ....lite.check import check
from ....sockets.addresses import SocketAddress
from ...handlers import HttpHandler
from ..io import CoroHttpIo
from .server import CoroHttpServer


##


class CoroHttpServerConnectionFdioHandler(SocketFdioHandler):
    def __init__(
            self,
            addr: SocketAddress,
            sock: socket.socket,
            handler: HttpHandler,
            *,
            read_size: int = 0x10000,
            write_size: int = 0x10000,
            log_handler: ta.Optional[ta.Callable[[CoroHttpServer, CoroHttpIo.AnyLogIo], None]] = None,
    ) -> None:
        check.state(not sock.getblocking())

        super().__init__(addr, sock)

        self._handler = handler
        self._read_size = read_size
        self._write_size = write_size
        self._log_handler = log_handler

        self._read_buf = ReadableListBuffer()
        self._write_buf: ta.Optional[IncrementalWriteBuffer] = None

        self._coro_srv = CoroHttpServer(
            addr,
            handler=self._handler,
        )
        self._srv_coro: ta.Optional[
            ta.Generator[
                CoroHttpIo.Io,
                ta.Optional[bytes],
                CoroHttpServer.CoroHandleResult,
            ],
        ] = self._coro_srv.coro_handle()

        self._cur_io: ta.Optional[CoroHttpIo.Io] = None
        self._next_io()

    #

    def _next_io(self) -> None:  # noqa
        coro = check.not_none(self._srv_coro)

        d: ta.Optional[bytes] = None
        o = self._cur_io
        while True:
            if o is None:
                try:
                    if d is not None:
                        o = coro.send(d)
                        d = None
                    else:
                        o = next(coro)
                except StopIteration:
                    self.close()
                    o = None
                    break

            if isinstance(o, CoroHttpIo.AnyLogIo):
                if self._log_handler is not None:
                    self._log_handler(self._coro_srv, o)
                o = None

            elif isinstance(o, CoroHttpIo.ReadIo):
                if (d := self._read_buf.read(o.sz)) is None:
                    break
                o = None

            elif isinstance(o, CoroHttpIo.ReadLineIo):
                if (d := self._read_buf.read_until(b'\n')) is None:
                    break
                o = None

            elif isinstance(o, CoroHttpIo.ReadUntilIo):
                # FIXME lol no max size, this is all getting thrown away anyawy
                if (d := self._read_buf.read_until(o.b)) is None:
                    break
                o = None

            elif isinstance(o, CoroHttpIo.WriteIo):
                check.none(self._write_buf)
                self._write_buf = IncrementalWriteBuffer(o.data, write_size=self._write_size)
                break

            else:
                raise TypeError(o)

        self._cur_io = o

    #

    def readable(self) -> bool:
        return True

    def writable(self) -> bool:
        return self._write_buf is not None

    #

    def on_readable(self) -> None:
        try:
            buf = check.not_none(self._sock).recv(self._read_size)
        except BlockingIOError:
            return
        except ConnectionResetError:
            self.close()
            return
        if not buf:
            self.close()
            return

        self._read_buf.feed(buf)

        if isinstance(self._cur_io, CoroHttpIo.AnyReadIo):
            self._next_io()

    def on_writable(self) -> None:
        check.isinstance(self._cur_io, CoroHttpIo.WriteIo)
        wb = check.not_none(self._write_buf)
        while wb.rem > 0:
            def send(d: bytes) -> int:
                try:
                    return check.not_none(self._sock).send(d)
                except ConnectionResetError:
                    self.close()
                    return 0
                except BlockingIOError:
                    return 0
            if not wb.write(send):
                break

        if wb.rem < 1:
            self._write_buf = None
            self._cur_io = None
            self._next_io()



from ... import dataclasses as dc
from ... import lang
from ... import marshal as msh
from .exprs import CanExpr
from .exprs import Expr
from .exprs import ExprBuilder
from .relations import CanRelation
from .relations import Relation
from .relations import RelationBuilder
from .stmts import Stmt


##


class Delete(Stmt, lang.Final):
    from_: Relation = dc.xfield() | msh.with_field_options(name='from')  # noqa
    where: Expr | None = dc.xfield(None, repr_fn=lang.opt_repr) | msh.with_field_options(omit_if=lang.is_none)


class DeleteBuilder(RelationBuilder, ExprBuilder):
    def delete(
            self,
            from_: CanRelation,
            where: CanExpr | None = None,
    ) -> Delete:
        return Delete(
            from_=self.relation(from_),
            where=self.expr(where) if where is not None else None,
        )



import abc
import functools
import os
import typing as ta

from .. import check
from .. import dataclasses as dc
from .. import lang


if ta.TYPE_CHECKING:
    import runpy

    from . import services

else:
    runpy = lang.proxy_import('runpy')

    services = lang.proxy_import('.services', __package__)


##


class Target(dc.Case):
    @classmethod
    def of(cls, obj: ta.Any) -> 'Target':
        if isinstance(obj, Target):
            return obj

        elif isinstance(obj, str):
            return NameTarget(obj)

        elif callable(obj):
            return FnTarget(obj)

        elif isinstance(obj, services.Service):
            return services.ServiceTarget(obj)

        elif isinstance(obj, services.Service.Config):
            return services.ServiceConfigTarget(obj)

        else:
            raise TypeError(obj)


class TargetRunner(lang.Abstract):
    @abc.abstractmethod
    def run(self) -> None:
        raise NotImplementedError


@functools.singledispatch
def target_runner_for(target: Target) -> TargetRunner:
    raise TypeError(target)


##


class FnTarget(Target):
    fn: ta.Callable[[], ta.Any]


class FnTargetRunner(TargetRunner, dc.Frozen):
    target: FnTarget

    def run(self) -> None:
        obj = self.target.fn()
        if obj is not None:
            tgt = Target.of(obj)
            tr = target_runner_for(tgt)
            tr.run()


@target_runner_for.register
def _(target: FnTarget) -> FnTargetRunner:
    return FnTargetRunner(target)


##


class NameTarget(Target):
    name: str

    @classmethod
    def for_obj(
            cls,
            obj: ta.Any,
            *,
            globals: ta.Mapping[str, ta.Any] | None = None,  # noqa
            no_module_name_lookup: bool = False,
    ) -> 'NameTarget':
        if globals is None:
            globals = obj.__globals__  # noqa
        if not no_module_name_lookup:
            mn = lang.get_real_module_name(globals)
        else:
            mn = globals['__name__']
        qn = obj.__qualname__
        return NameTarget('.'.join([mn, qn]))


class NameTargetRunner(TargetRunner, dc.Frozen):
    target: NameTarget

    def run(self) -> None:
        name = self.target.name
        if lang.can_import(name):
            runpy._run_module_as_main(name)  # type: ignore  # noqa
        else:
            obj = lang.import_attr(self.target.name)
            tgt = Target.of(obj)
            tr = target_runner_for(tgt)
            tr.run()


@target_runner_for.register
def _(target: NameTarget) -> NameTargetRunner:
    return NameTargetRunner(target)


##


class ExecTarget(Target):
    cmd: ta.Sequence[str] = dc.xfield(coerce=check.of_not_isinstance(str))

    cwd: str | None = None


class ExecTargetRunner(TargetRunner, dc.Frozen):
    target: ExecTarget

    def run(self) -> None:
        if (cwd := self.target.cwd) is not None:
            os.chdir(os.path.expanduser(cwd))

        os.execl(*self.target.cmd)


@target_runner_for.register
def _(target: ExecTarget) -> ExecTargetRunner:
    return ExecTargetRunner(target)



import contextvars
import typing as ta

from .. import asgi


##


SCOPE: contextvars.ContextVar[asgi.Scope] = contextvars.ContextVar('scope')


##


BaseServerUrl = ta.NewType('BaseServerUrl', str)


BASE_SERVER_URL: contextvars.ContextVar[BaseServerUrl] = contextvars.ContextVar('base_server_url')


def url_for(s: str) -> str:
    return BASE_SERVER_URL.get() + s
